# /*##########################################################################
#
# Copyright (c) 2016-2021 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
__authors__ = ["P. Knobel"]
__license__ = "MIT"
__date__ = "05/12/2016"

import os
import tempfile
import unittest

import numpy

from silx.gui import qt
from silx.gui.data import ArrayTableWidget
from silx.gui.data.ArrayTableModel import ArrayTableModel
from silx.gui.utils.testutils import TestCaseQt

import h5py


class TestArrayWidget(TestCaseQt):
    """Basic test for ArrayTableWidget with a numpy array"""
    def setUp(self):
        super(TestArrayWidget, self).setUp()
        self.aw = ArrayTableWidget.ArrayTableWidget()

    def tearDown(self):
        del self.aw
        super(TestArrayWidget, self).tearDown()

    def testShow(self):
        """test for errors"""
        self.aw.show()
        self.qWaitForWindowExposed(self.aw)

    def testSetData0D(self):
        a = 1
        self.aw.setArrayData(a)
        b = self.aw.getData(copy=True)

        self.assertTrue(numpy.array_equal(a, b))

        # scalar/0D data has no frame index
        self.assertEqual(len(self.aw.model._index), 0)
        # and no perspective
        self.assertEqual(len(self.aw.model._perspective), 0)

    def testSetData1D(self):
        a = [1, 2]
        self.aw.setArrayData(a)
        b = self.aw.getData(copy=True)

        self.assertTrue(numpy.array_equal(a, b))

        # 1D data has no frame index
        self.assertEqual(len(self.aw.model._index), 0)
        # and no perspective
        self.assertEqual(len(self.aw.model._perspective), 0)

    def testSetData4D(self):
        a = numpy.reshape(numpy.linspace(0.213, 1.234, 1250),
                          (5, 5, 5, 10))
        self.aw.setArrayData(a)

        # default perspective (0, 1)
        self.assertEqual(list(self.aw.model._perspective),
                         [0, 1])
        self.aw.setPerspective((1, 3))
        self.assertEqual(list(self.aw.model._perspective),
                         [1, 3])

        b = self.aw.getData(copy=True)
        self.assertTrue(numpy.array_equal(a, b))

        # 4D data has a 2-tuple as frame index
        self.assertEqual(len(self.aw.model._index), 2)
        # default index is (0, 0)
        self.assertEqual(list(self.aw.model._index),
                         [0, 0])
        self.aw.setFrameIndex((3, 1))

        self.assertEqual(list(self.aw.model._index),
                         [3, 1])

    def testColors(self):
        a = numpy.arange(256, dtype=numpy.uint8)
        self.aw.setArrayData(a)

        bgcolor = numpy.empty(a.shape + (3,), dtype=numpy.uint8)
        # Black & white palette
        bgcolor[..., 0] = a
        bgcolor[..., 1] = a
        bgcolor[..., 2] = a

        fgcolor = numpy.bitwise_xor(bgcolor, 255)

        self.aw.setArrayColors(bgcolor, fgcolor)

        # test colors are as expected in model
        for i in range(256):
            # all RGB channels for BG equal to data value
            self.assertEqual(
                self.aw.model.data(self.aw.model.index(0, i),
                                   role=qt.Qt.BackgroundRole),
                qt.QColor(i, i, i),
                "Unexpected background color"
            )

            # all RGB channels for FG equal to XOR(data value, 255)
            self.assertEqual(
                self.aw.model.data(self.aw.model.index(0, i),
                                   role=qt.Qt.ForegroundRole),
                qt.QColor(i ^ 255, i ^ 255, i ^ 255),
                "Unexpected text color"
            )

        # test colors are reset to None when a new data array is loaded
        # with different shape
        self.aw.setArrayData(numpy.arange(300))

        for i in range(300):
            # all RGB channels for BG equal to data value
            self.assertIsNone(
                self.aw.model.data(self.aw.model.index(0, i),
                                   role=qt.Qt.BackgroundRole))

    def testDefaultFlagNotEditable(self):
        """editable should be False by default, in setArrayData"""
        self.aw.setArrayData([[0]])
        idx = self.aw.model.createIndex(0, 0)
        # model is editable
        self.assertFalse(
                self.aw.model.flags(idx) & qt.Qt.ItemIsEditable)

    def testFlagEditable(self):
        self.aw.setArrayData([[0]], editable=True)
        idx = self.aw.model.createIndex(0, 0)
        # model is editable
        self.assertTrue(
                self.aw.model.flags(idx) & qt.Qt.ItemIsEditable)

    def testFlagNotEditable(self):
        self.aw.setArrayData([[0]], editable=False)
        idx = self.aw.model.createIndex(0, 0)
        # model is editable
        self.assertFalse(
                self.aw.model.flags(idx) & qt.Qt.ItemIsEditable)

    def testReferenceReturned(self):
        """when setting the data with copy=False and
        retrieving it with getData(copy=False), we should recover
        the same original object.
        """
        # n-D (n >=2)
        a0 = numpy.reshape(numpy.linspace(0.213, 1.234, 1000),
                           (10, 10, 10))
        self.aw.setArrayData(a0, copy=False)
        a1 = self.aw.getData(copy=False)

        self.assertIs(a0, a1)

        # 1D
        b0 = numpy.linspace(0.213, 1.234, 1000)
        self.aw.setArrayData(b0, copy=False)
        b1 = self.aw.getData(copy=False)
        self.assertIs(b0, b1)

    def testClipping(self):
        """Test clipping of large arrays"""
        self.aw.show()
        self.qWaitForWindowExposed(self.aw)

        data = numpy.arange(ArrayTableModel.MAX_NUMBER_OF_SECTIONS + 10)

        for shape in [(1, -1), (-1, 1)]:
            with self.subTest(shape=shape):
                self.aw.setArrayData(data.reshape(shape), editable=True)
                self.qapp.processEvents()


class TestH5pyArrayWidget(TestCaseQt):
    """Basic test for ArrayTableWidget with a dataset.

    Test flags, for dataset open in read-only or read-write modes"""
    def setUp(self):
        super(TestH5pyArrayWidget, self).setUp()
        self.aw = ArrayTableWidget.ArrayTableWidget()
        self.data = numpy.reshape(numpy.linspace(0.213, 1.234, 1000),
                                  (10, 10, 10))
        # create an h5py file with a dataset
        self.tempdir = tempfile.mkdtemp()
        self.h5_fname = os.path.join(self.tempdir, "array.h5")
        h5f = h5py.File(self.h5_fname, mode='w')
        h5f["my_array"] = self.data
        h5f["my_scalar"] = 3.14
        h5f["my_1D_array"] = numpy.array(numpy.arange(1000))
        h5f.close()

    def tearDown(self):
        del self.aw
        os.unlink(self.h5_fname)
        os.rmdir(self.tempdir)
        super(TestH5pyArrayWidget, self).tearDown()

    def testShow(self):
        self.aw.show()
        self.qWaitForWindowExposed(self.aw)

    def testReadOnly(self):
        """Open H5 dataset in read-only mode, ensure the model is not editable."""
        h5f = h5py.File(self.h5_fname, "r")
        a = h5f["my_array"]
        # ArrayTableModel relies on following condition
        self.assertTrue(a.file.mode == "r")

        self.aw.setArrayData(a, copy=False, editable=True)

        self.assertIsInstance(a, h5py.Dataset)   # simple sanity check
        # internal representation must be a reference to original data (copy=False)
        self.assertIsInstance(self.aw.model._array, h5py.Dataset)
        self.assertTrue(self.aw.model._array.file.mode == "r")

        b = self.aw.getData()
        self.assertTrue(numpy.array_equal(self.data, b))

        # model must have detected read-only dataset and disabled editing
        self.assertFalse(self.aw.model._editable)
        idx = self.aw.model.createIndex(0, 0)
        self.assertFalse(
                 self.aw.model.flags(idx) & qt.Qt.ItemIsEditable)

        # force editing read-only datasets raises IOError
        self.assertRaises(IOError, self.aw.model.setData,
                          idx, 123.4, role=qt.Qt.EditRole)
        h5f.close()

    def testReadWrite(self):
        h5f = h5py.File(self.h5_fname, "r+")
        a = h5f["my_array"]
        self.assertTrue(a.file.mode == "r+")

        self.aw.setArrayData(a, copy=False, editable=True)
        b = self.aw.getData(copy=False)
        self.assertTrue(numpy.array_equal(self.data, b))

        idx = self.aw.model.createIndex(0, 0)
        # model is editable
        self.assertTrue(
                self.aw.model.flags(idx) & qt.Qt.ItemIsEditable)
        h5f.close()

    def testSetData0D(self):
        h5f = h5py.File(self.h5_fname, "r+")
        a = h5f["my_scalar"]
        self.aw.setArrayData(a)
        b = self.aw.getData(copy=True)

        self.assertTrue(numpy.array_equal(a, b))

        h5f.close()

    def testSetData1D(self):
        h5f = h5py.File(self.h5_fname, "r+")
        a = h5f["my_1D_array"]
        self.aw.setArrayData(a)
        b = self.aw.getData(copy=True)

        self.assertTrue(numpy.array_equal(a, b))

        h5f.close()

    def testReferenceReturned(self):
        """when setting the data with copy=False and
        retrieving it with getData(copy=False), we should recover
        the same original object.

        This only works for array with at least 2D. For 1D and 0D
        arrays, a view is created at some point, which  in the case
        of an hdf5 dataset creates a copy."""
        h5f = h5py.File(self.h5_fname, "r+")

        # n-D
        a0 = h5f["my_array"]
        self.aw.setArrayData(a0, copy=False)
        a1 = self.aw.getData(copy=False)
        self.assertIs(a0, a1)

        # 1D
        b0 = h5f["my_1D_array"]
        self.aw.setArrayData(b0, copy=False)
        b1 = self.aw.getData(copy=False)
        self.assertIs(b0, b1)

        h5f.close()



# /*##########################################################################
#
# Copyright (c) 2016-2018 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""This module provides combination of statistics as single operation.

For now it provides min/max (and optionally positive min) and indices
of first occurrences (i.e., argmin/argmax) in a single pass.
"""

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "24/04/2018"

cimport cython
from .math_compatibility cimport isnan, isfinite, INFINITY


import numpy


# All supported types
ctypedef fused _number:
    float
    double
    long double
    signed char
    signed short
    signed int
    signed long
    signed long long
    unsigned char
    unsigned short
    unsigned int
    unsigned long
    unsigned long long

# All supported floating types:
# cython.floating + long double
ctypedef fused _floating:
    float
    double
    long double


class _MinMaxResult(object):
    """Object storing result from :func:`min_max`"""

    def __init__(self, minimum, min_pos, maximum,
                 argmin, argmin_pos, argmax):
        self._minimum = minimum
        self._min_positive = min_pos
        self._maximum = maximum

        self._argmin = argmin
        self._argmin_positive = argmin_pos
        self._argmax = argmax

    minimum = property(
        lambda self: self._minimum,
        doc="Minimum value of the array")
    maximum = property(
        lambda self: self._maximum,
        doc="Maximum value of the array")

    argmin = property(
        lambda self: self._argmin,
        doc="Index of the first occurrence of the minimum value")
    argmax = property(
        lambda self: self._argmax,
        doc="Index of the first occurrence of the maximum value")

    min_positive = property(
        lambda self: self._min_positive,
        doc="""Strictly positive minimum value

        It is None if no value is strictly positive.
        """)
    argmin_positive = property(
        lambda self: self._argmin_positive,
        doc="""Index of the strictly positive minimum value.

        It is None if no value is strictly positive.
        It is the index of the first occurrence.""")

    def __getitem__(self, key):
        if key == 0:
            return self.minimum
        elif key == 1:
            return self.maximum
        else:
            raise IndexError("Index out of range")


@cython.initializedcheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
def _min_max(_number[::1] data, bint min_positive=False):
    """:func:`min_max` implementation including infinite values

    See :func:`min_max` for documentation.
    """
    cdef:
        _number value, minimum, min_pos, maximum
        unsigned int length
        unsigned int index = 0
        unsigned int min_index = 0
        unsigned int min_pos_index = 0
        unsigned int max_index = 0

    length = len(data)

    if length == 0:
        raise ValueError('Zero-size array')

    with nogil:
        # Init starting values
        value = data[0]
        minimum = value
        maximum = value
        if min_positive and value > 0:
            min_pos = value
        else:
            min_pos = 0

        if _number in _floating:
            # For floating, loop until first not NaN value
            for index in range(length):
                value = data[index]
                if not isnan(value):
                    minimum = value
                    min_index = index
                    maximum = value
                    max_index = index
                    break

        if not min_positive:
            for index in range(index, length):
                value = data[index]
                if value > maximum:
                    maximum = value
                    max_index = index
                elif value < minimum:
                    minimum = value
                    min_index = index

        else:
            # Loop until min_pos is defined
            for index in range(index, length):
                value = data[index]
                if value > maximum:
                    maximum = value
                    max_index = index
                elif value < minimum:
                    minimum = value
                    min_index = index

                if value > 0:
                    min_pos = value
                    min_pos_index = index
                    break

            # Loop until the end
            for index in range(index + 1, length):
                value = data[index]
                if value > maximum:
                    maximum = value
                    max_index = index
                else:
                    if value < minimum:
                        minimum = value
                        min_index = index

                    if 0 < value < min_pos:
                        min_pos = value
                        min_pos_index = index

    return _MinMaxResult(minimum,
                         min_pos if min_pos > 0 else None,
                         maximum,
                         min_index,
                         min_pos_index if min_pos > 0 else None,
                         max_index)


@cython.initializedcheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
def _finite_min_max(_floating[::1] data, bint min_positive=False):
    """:func:`min_max` implementation for floats skipping infinite values

    See :func:`min_max` for documentation.
    """
    cdef:
        _floating value, minimum, min_pos, maximum
        unsigned int length
        unsigned int index = 0
        unsigned int min_index = 0
        unsigned int min_pos_index = 0
        unsigned int max_index = 0

    length = len(data)

    if length == 0:
        raise ValueError('Zero-size array')

    with nogil:
        minimum = INFINITY
        maximum = -INFINITY
        min_pos = INFINITY

        if not min_positive:
            for index in range(length):
                value = data[index]
                if isfinite(value):
                    if value > maximum:
                        maximum = value
                        max_index = index
                    if value < minimum:
                        minimum = value
                        min_index = index

        else:
            for index in range(index, length):
                value = data[index]
                if isfinite(value):
                    if value > maximum:
                        maximum = value
                        max_index = index
                    if value < minimum:
                        minimum = value
                        min_index = index

                    if 0. < value < min_pos:
                        min_pos = value
                        min_pos_index = index

    return _MinMaxResult(minimum if isfinite(minimum) else None,
                         min_pos if isfinite(min_pos) else None,
                         maximum if isfinite(maximum) else None,
                         min_index if isfinite(minimum) else None,
                         min_pos_index if isfinite(min_pos) else None,
                         max_index if isfinite(maximum) else None)


def min_max(data not None, bint min_positive=False, bint finite=False):
    """Returns min, max and optionally strictly positive min of data.

    It also computes the indices of first occurrence of min/max.

    NaNs are ignored while computing min/max unless all data is NaNs,
    in which case returned min/max are NaNs.

    The result data type is that of the input data, except for the following cases.
    For input using non-native bytes order, the result is returned as native
    floating-point or integers. For input using 16-bits floating-point,
    the result is returned as 32-bits floating-point.

    Examples:

    >>> import numpy
    >>> data = numpy.arange(10)

    Usage as a function returning min and max:

    >>> min_, max_ = min_max(data)

    Usage as a function returning a result object to access all information:

    >>> result = min_max(data)  # Do not get positive min
    >>> result.minimum, result.argmin
    0, 0
    >>> result.maximum, result.argmax
    9, 10
    >>> result.min_positive, result.argmin_positive  # Not computed
    None, None

    Getting strictly positive min information:

    >>> result = min_max(data, min_positive=True)
    >>> result.min_positive, result.argmin_positive  # Computed
    1, 1

    If *finite* is True, min/max information is computed only from finite data.
    Then, all result fields (include minimum and maximum) can be None
    when all data is infinity or NaN.

    :param data: Array-like dataset
    :param bool min_positive: True to compute the positive min and argmin
                              Default: False.
    :param bool finite: True to compute min/max from finite data only
                        Default: False.
    :returns: An object with minimum, maximum and min_positive attributes
              and the indices of first occurrence in the flattened data:
              argmin, argmax and argmin_positive attributes.
              If all data is <= 0 or min_positive argument is False, then
              min_positive and argmin_positive are None.
    :raises: ValueError if data is empty
    """
    data = numpy.array(data, copy=False)
    native_endian_dtype = data.dtype.newbyteorder('N')
    if native_endian_dtype.kind == 'f' and native_endian_dtype.itemsize == 2:
        # Use native float32 instead of float16
        native_endian_dtype = "=f4"
    data = numpy.ascontiguousarray(data, dtype=native_endian_dtype).ravel()
    if finite and data.dtype.kind == 'f':
        return _finite_min_max(data, min_positive)
    else:
        return _min_max(data, min_positive)



# /*##########################################################################
#
# Copyright (c) 2016-2018 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""This module defines two main classes:

    - :class:`FrameBrowser`: a widget with 4 buttons (first, previous, next,
      last) to browse between frames and a text entry to access a specific frame
      by typing it's number)
    - :class:`HorizontalSliderWithBrowser`: a FrameBrowser with an additional
      slider. This class inherits :class:`qt.QAbstractSlider`.

"""
from silx.gui import qt
from silx.gui import icons
from silx.utils import deprecation

__authors__ = ["V.A. Sole", "P. Knobel"]
__license__ = "MIT"
__date__ = "16/01/2017"


class FrameBrowser(qt.QWidget):
    """Frame browser widget, with 4 buttons/icons and a line edit to provide
    a way of selecting a frame index in a stack of images.

    .. image:: img/FrameBrowser.png

    It can be used in more generic case to select an integer within a range.

    :param QWidget parent: Parent widget
    :param int n: Number of frames. This will set the range
        of frame indices to 0--n-1.
        If None, the range is initialized to the default QSlider range (0--99).
    """

    sigIndexChanged = qt.pyqtSignal(object)

    def __init__(self, parent=None, n=None):
        qt.QWidget.__init__(self, parent)

        # Use the font size as the icon size to avoid to create bigger buttons
        fontMetric = self.fontMetrics()
        iconSize = qt.QSize(fontMetric.height(), fontMetric.height())

        self.mainLayout = qt.QHBoxLayout(self)
        self.mainLayout.setContentsMargins(0, 0, 0, 0)
        self.mainLayout.setSpacing(0)
        self.firstButton = qt.QPushButton(self)
        self.firstButton.setIcon(icons.getQIcon("first"))
        self.firstButton.setIconSize(iconSize)
        self.previousButton = qt.QPushButton(self)
        self.previousButton.setIcon(icons.getQIcon("previous"))
        self.previousButton.setIconSize(iconSize)
        self._lineEdit = qt.QLineEdit(self)

        self._label = qt.QLabel(self)
        self.nextButton = qt.QPushButton(self)
        self.nextButton.setIcon(icons.getQIcon("next"))
        self.nextButton.setIconSize(iconSize)
        self.lastButton = qt.QPushButton(self)
        self.lastButton.setIcon(icons.getQIcon("last"))
        self.lastButton.setIconSize(iconSize)

        self.mainLayout.addWidget(self.firstButton)
        self.mainLayout.addWidget(self.previousButton)
        self.mainLayout.addWidget(self._lineEdit)
        self.mainLayout.addWidget(self._label)
        self.mainLayout.addWidget(self.nextButton)
        self.mainLayout.addWidget(self.lastButton)

        if n is None:
            first = qt.QSlider().minimum()
            last = qt.QSlider().maximum()
        else:
            first, last = 0, n

        self._lineEdit.setFixedWidth(self._lineEdit.fontMetrics().boundingRect('%05d' % last).width())
        validator = qt.QIntValidator(first, last, self._lineEdit)
        self._lineEdit.setValidator(validator)
        self._lineEdit.setText("%d" % first)
        self._label.setText("of %d" % last)

        self._index = first
        """0-based index"""

        self.firstButton.clicked.connect(self._firstClicked)
        self.previousButton.clicked.connect(self._previousClicked)
        self.nextButton.clicked.connect(self._nextClicked)
        self.lastButton.clicked.connect(self._lastClicked)
        self._lineEdit.editingFinished.connect(self._textChangedSlot)

    def lineEdit(self):
        """Returns the line edit provided by this widget.

        :rtype: qt.QLineEdit
        """
        return self._lineEdit

    def limitWidget(self):
        """Returns the widget displaying axes limits.

        :rtype: qt.QLabel
        """
        return self._label

    def _firstClicked(self):
        """Select first/lowest frame number"""
        self.setValue(self.getRange()[0])

    def _previousClicked(self):
        """Select previous frame number"""
        self.setValue(self.getValue() - 1)

    def _nextClicked(self):
        """Select next frame number"""
        self.setValue(self.getValue() + 1)

    def _lastClicked(self):
        """Select last/highest frame number"""
        self.setValue(self.getRange()[1])

    def _textChangedSlot(self):
        """Select frame number typed in the line edit widget"""
        txt = self._lineEdit.text()
        if not len(txt):
            self._lineEdit.setText("%d" % self._index)
            return
        new_value = int(txt)
        if new_value == self._index:
            return
        ddict = {
            "event": "indexChanged",
            "old": self._index,
            "new": new_value,
            "id": id(self)
        }
        self._index = new_value
        self.sigIndexChanged.emit(ddict)

    def getRange(self):
        """Returns frame range

        :return: (first_index, last_index)
        """
        validator = self.lineEdit().validator()
        return validator.bottom(), validator.top()

    def setRange(self, first, last):
        """Set minimum and maximum frame indices.

        Initialize the frame index to *first*.
        Update the label text to *" limits: first, last"*

        :param int first: Minimum frame index
        :param int last: Maximum frame index"""
        bottom = min(first, last)
        top = max(first, last)
        self._lineEdit.validator().setTop(top)
        self._lineEdit.validator().setBottom(bottom)
        self.setValue(bottom)

        # Update limits
        self._label.setText(" limits: %d, %d " % (bottom, top))

    @deprecation.deprecated(replacement="FrameBrowser.setRange",
                            since_version="0.8")
    def setLimits(self, first, last):
        return self.setRange(first, last)

    def setNFrames(self, nframes):
        """Set minimum=0 and maximum=nframes-1 frame numbers.

        Initialize the frame index to 0.
        Update the label text to *"1 of nframes"*

        :param int nframes: Number of frames"""
        top = nframes - 1
        self.setRange(0, top)
        # display 1-based index in label
        self._label.setText(" of %d " % top)

    @deprecation.deprecated(replacement="FrameBrowser.getValue",
                            since_version="0.8")
    def getCurrentIndex(self):
        return self._index

    def getValue(self):
        """Return current frame index"""
        return self._index

    def setValue(self, value):
        """Set 0-based frame index

        Value is clipped to current range.

        :param int value: Frame number"""
        bottom = self.lineEdit().validator().bottom()
        top = self.lineEdit().validator().top()
        value = int(value)

        if value < bottom:
            value = bottom
        elif value > top:
            value = top

        self._lineEdit.setText("%d" % value)
        self._textChangedSlot()


class HorizontalSliderWithBrowser(qt.QAbstractSlider):
    """
    Slider widget combining a :class:`QSlider` and a :class:`FrameBrowser`.

    .. image:: img/HorizontalSliderWithBrowser.png

    The data model is an integer within a range.

    The default value is the default :class:`QSlider` value (0),
    and the default range is the default QSlider range (0 -- 99)

    The signal emitted when the value is changed is the usual QAbstractSlider
    signal :attr:`valueChanged`. The signal carries the value (as an integer).

    :param QWidget parent: Optional parent widget
    """
    def __init__(self, parent=None):
        qt.QAbstractSlider.__init__(self, parent)
        self.setOrientation(qt.Qt.Horizontal)

        self.mainLayout = qt.QHBoxLayout(self)
        self.mainLayout.setContentsMargins(0, 0, 0, 0)
        self.mainLayout.setSpacing(2)

        self._slider = qt.QSlider(self)
        self._slider.setOrientation(qt.Qt.Horizontal)

        self._browser = FrameBrowser(self)

        self.mainLayout.addWidget(self._slider, 1)
        self.mainLayout.addWidget(self._browser)

        self._slider.valueChanged[int].connect(self._sliderSlot)
        self._browser.sigIndexChanged.connect(self._browserSlot)

    def lineEdit(self):
        """Returns the line edit provided by this widget.

        :rtype: qt.QLineEdit
        """
        return self._browser.lineEdit()

    def limitWidget(self):
        """Returns the widget displaying axes limits.

        :rtype: qt.QLabel
        """
        return self._browser.limitWidget()

    def setMinimum(self, value):
        """Set minimum value

        :param int value: Minimum value"""
        self._slider.setMinimum(value)
        maximum = self._slider.maximum()
        self._browser.setRange(value, maximum)

    def setMaximum(self, value):
        """Set maximum value

        :param int value: Maximum value
        """
        self._slider.setMaximum(value)
        minimum = self._slider.minimum()
        self._browser.setRange(minimum, value)

    def setRange(self, first, last):
        """Set minimum/maximum values

        :param int first: Minimum value
        :param int last: Maximum value"""
        self._slider.setRange(first, last)
        self._browser.setRange(first, last)

    def _sliderSlot(self, value):
        """Emit selected value when slider is activated
        """
        self._browser.setValue(value)
        self.valueChanged.emit(value)

    def _browserSlot(self, ddict):
        """Emit selected value when browser state is changed"""
        self._slider.setValue(ddict['new'])

    def setValue(self, value):
        """Set value

        :param int value: value"""
        self._slider.setValue(value)
        self._browser.setValue(value)

    def value(self):
        """Get selected value"""
        return self._slider.value()



# /*##########################################################################
#
# Copyright (c) 2016-2018 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""This package provides a set of Qt widgets for plotting curves and images.

The plotting API is inherited from the `PyMca <http://pymca.sourceforge.net/>`_
plot API and is mostly compatible with it.

Those widgets supports interaction (e.g., zoom, pan, selections).

List of Qt widgets:

.. currentmodule:: silx.gui.plot

- :mod:`.PlotWidget`: A widget displaying a single plot.
- :mod:`.PlotWindow`: A :mod:`.PlotWidget` with a configurable set of tools.
- :class:`.Plot1D`: A widget with tools for curves.
- :class:`.Plot2D`: A widget with tools for images.
- :class:`.ScatterView`: A widget with tools for scatter plot.
- :class:`.ImageView`: A widget with tools for images and a side histogram.
- :class:`.StackView`: A widget with tools for a stack of images.

By default, those widget are using matplotlib_.
They can optionally use a faster OpenGL-based rendering (beta feature),
which is enabled by setting the ``backend`` argument to ``'gl'``
when creating the widgets (See :class:`.PlotWidget`).

.. note::

    This package depends on matplotlib_.
    The OpenGL backend further depends on
    `PyOpenGL <http://pyopengl.sourceforge.net/>`_ and OpenGL >= 2.1.

.. _matplotlib: http://matplotlib.org/
"""

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "03/05/2017"


from .PlotWidget import PlotWidget  # noqa
from .PlotWindow import PlotWindow, Plot1D, Plot2D  # noqa
from .items.axis import TickMode
from .ImageView import ImageView  # noqa
from .StackView import StackView  # noqa
from .ScatterView import ScatterView  # noqa

__all__ = ['ImageView', 'PlotWidget', 'PlotWindow', 'Plot1D', 'Plot2D',
           'StackView', 'ScatterView', 'TickMode']



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2017-2022 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""This script illustrates the update of a :class:`~silx.gui.plot.Plot2D`
widget from a gevent coroutine.
"""

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "05/09/2017"


import threading
import time
import gevent
import logging

import numpy

from silx.gui import qt
from silx.gui.utils import concurrent
from silx.gui.plot import Plot2D

_logger = logging.getLogger(__name__)

Nx = 150
Ny = 50


def process_gevent():
    """Process gevent in case of QTimer triggering it."""
    try:
        gevent.sleep(0.01)
    except Exception:
        _logger.critical("Uncaught exception from gevent", exc_info=True)


def update_image(plot2d):
    """Update the image of a :class:`~sil.gui.plot.Plot2D`

    :param plot2d: The Plot2D to update."""

    pos = {'x0': 0, 'y0': 0}
    while True:
        gevent.sleep(0.01)

        # Create image
        # width of peak
        sigma_x = 0.15
        sigma_y = 0.25
        # x and y positions
        x = numpy.linspace(-1.5, 1.5, Nx)
        y = numpy.linspace(-1.0, 1.0, Ny)
        xv, yv = numpy.meshgrid(x, y)
        signal = numpy.exp(- ((xv - pos['x0']) ** 2 / sigma_x ** 2
                                + (yv - pos['y0']) ** 2 / sigma_y ** 2))
        # add noise
        signal += 0.3 * numpy.random.random(size=signal.shape)
        # random walk of center of peak ('drift')
        pos['x0'] += 0.05 * (numpy.random.random() - 0.5)
        pos['y0'] += 0.05 * (numpy.random.random() - 0.5)

        plot2d.addImage(signal, resetzoom=False)


def main():
    global app
    app = qt.QApplication([])

    gevent_timer = qt.QTimer()
    gevent_timer.start(10)
    gevent_timer.timeout.connect(process_gevent)

    # Create a Plot2D, set its limits and display it
    plot2d = Plot2D()
    plot2d.getIntensityHistogramAction().setVisible(True)
    plot2d.setLimits(0, Nx, 0, Ny)
    plot2d.getDefaultColormap().setVRange(0., 1.5)
    plot2d.show()

    # Create the thread that calls submitToQtMainThread
    updater = gevent.spawn(update_image, plot2d)

    app.exec()

    updater.kill()  # Stop updating the plot
    updater.join()


if __name__ == '__main__':
    main()



# /*##########################################################################
#
# Copyright (c) 2018 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "28/06/2018"


import unittest
import numpy

from silx.gui import qt
from silx.utils.testutils import ParametricTestCase
from silx.gui.utils.testutils import TestCaseQt
from silx.gui.plot import PlotWindow
from silx.gui.plot.tools.profile import manager
from silx.gui.plot.tools.profile import core
from silx.gui.plot.tools.profile import rois


class TestScatterProfileToolBar(TestCaseQt, ParametricTestCase):
    """Tests for ScatterProfileToolBar class"""

    def setUp(self):
        super(TestScatterProfileToolBar, self).setUp()
        self.plot = PlotWindow()

        self.manager = manager.ProfileManager(plot=self.plot)
        self.manager.setItemType(scatter=True)
        self.manager.setActiveItemTracking(True)

        self.plot.show()
        self.qWaitForWindowExposed(self.plot)

    def tearDown(self):
        del self.manager
        self.qapp.processEvents()
        self.plot.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.plot.close()
        del self.plot
        super(TestScatterProfileToolBar, self).tearDown()

    def testHorizontalProfile(self):
        """Test ScatterProfileToolBar horizontal profile"""

        roiManager = self.manager.getRoiManager()

        # Add a scatter plot
        self.plot.addScatter(
            x=(0., 1., 1., 0.), y=(0., 0., 1., 1.), value=(0., 1., 2., 3.))
        self.plot.resetZoom(dataMargins=(.1, .1, .1, .1))
        self.qapp.processEvents()

        # Set a ROI profile
        roi = rois.ProfileScatterHorizontalLineROI()
        roi.setPosition(0.5)
        roi.setNPoints(8)
        roiManager.addRoi(roi)

        # Wait for async interpolator init
        for _ in range(20):
            self.qWait(200)
            if not self.manager.hasPendingOperations():
                break
        self.qapp.processEvents()

        window = roi.getProfileWindow()
        self.assertIsNotNone(window)
        data = window.getProfile()
        self.assertIsInstance(data, core.CurveProfileData)
        self.assertEqual(len(data.coords), 8)

        # Check that profile has same limits than Plot
        xLimits = self.plot.getXAxis().getLimits()
        self.assertEqual(data.coords[0], xLimits[0])
        self.assertEqual(data.coords[-1], xLimits[1])

        # Clear the profile
        self.manager.clearProfile()
        self.qapp.processEvents()
        self.assertIsNone(roi.getProfileWindow())

    def testVerticalProfile(self):
        """Test ScatterProfileToolBar vertical profile"""

        roiManager = self.manager.getRoiManager()

        # Add a scatter plot
        self.plot.addScatter(
            x=(0., 1., 1., 0.), y=(0., 0., 1., 1.), value=(0., 1., 2., 3.))
        self.plot.resetZoom(dataMargins=(.1, .1, .1, .1))
        self.qapp.processEvents()

        # Set a ROI profile
        roi = rois.ProfileScatterVerticalLineROI()
        roi.setPosition(0.5)
        roi.setNPoints(8)
        roiManager.addRoi(roi)

        # Wait for async interpolator init
        for _ in range(10):
            self.qWait(200)
            if not self.manager.hasPendingOperations():
                break

        window = roi.getProfileWindow()
        self.assertIsNotNone(window)
        data = window.getProfile()
        self.assertIsInstance(data, core.CurveProfileData)
        self.assertEqual(len(data.coords), 8)

        # Check that profile has same limits than Plot
        yLimits = self.plot.getYAxis().getLimits()
        self.assertEqual(data.coords[0], yLimits[0])
        self.assertEqual(data.coords[-1], yLimits[1])

        # Check that profile limits are updated when changing limits
        self.plot.getYAxis().setLimits(yLimits[0] + 1, yLimits[1] + 10)

        # Wait for async interpolator init
        for _ in range(10):
            self.qWait(200)
            if not self.manager.hasPendingOperations():
                break

        yLimits = self.plot.getYAxis().getLimits()
        data = window.getProfile()
        self.assertEqual(data.coords[0], yLimits[0])
        self.assertEqual(data.coords[-1], yLimits[1])

        # Clear the profile
        self.manager.clearProfile()
        self.qapp.processEvents()
        self.assertIsNone(roi.getProfileWindow())

    def testLineProfile(self):
        """Test ScatterProfileToolBar line profile"""

        roiManager = self.manager.getRoiManager()

        # Add a scatter plot
        self.plot.addScatter(
            x=(0., 1., 1., 0.), y=(0., 0., 1., 1.), value=(0., 1., 2., 3.))
        self.plot.resetZoom(dataMargins=(.1, .1, .1, .1))
        self.qapp.processEvents()

        # Set a ROI profile
        roi = rois.ProfileScatterLineROI()
        roi.setEndPoints(numpy.array([0., 0.]), numpy.array([1., 1.]))
        roi.setNPoints(8)
        roiManager.addRoi(roi)

        # Wait for async interpolator init
        for _ in range(10):
            self.qWait(200)
            if not self.manager.hasPendingOperations():
                break

        window = roi.getProfileWindow()
        self.assertIsNotNone(window)
        data = window.getProfile()
        self.assertIsInstance(data, core.CurveProfileData)
        self.assertEqual(len(data.coords), 8)



# /*##########################################################################
# Copyright (C) 2016-2022 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ############################################################################*/
"""Tests for dicttoh5 module"""

__authors__ = ["P. Knobel"]
__license__ = "MIT"
__date__ = "17/01/2018"


from collections import defaultdict, OrderedDict
from copy import deepcopy
from io import BytesIO
import os
import tempfile
import unittest

import h5py
import numpy
try:
    import pint
except ImportError:
    pint = None
import pytest

from silx.utils.testutils import LoggingValidator

from ..configdict import ConfigDict
from .. import dictdump
from ..dictdump import dicttoh5, dicttojson, dump
from ..dictdump import h5todict, load
from ..dictdump import logger as dictdump_logger
from ..utils import is_link
from ..utils import h5py_read_dataset


@pytest.fixture
def tmp_h5py_file():
    with BytesIO() as buffer:
        with h5py.File(buffer, mode="w") as h5file:
            yield h5file


def tree():
    """Tree data structure as a recursive nested dictionary"""
    return defaultdict(tree)


inhabitants = 160215

city_attrs = tree()
city_attrs["Europe"]["France"]["Grenoble"]["area"] = "18.44 km2"
city_attrs["Europe"]["France"]["Grenoble"]["inhabitants"] = inhabitants
city_attrs["Europe"]["France"]["Grenoble"]["coordinates"] = [45.1830, 5.7196]
city_attrs["Europe"]["France"]["Tourcoing"]["area"]

ext_attrs = tree()
ext_attrs["ext_group"]["dataset"] = 10
ext_filename = "ext.h5"

link_attrs = tree()
link_attrs["links"]["group"]["dataset"] = 10
link_attrs["links"]["group"]["relative_softlink"] = h5py.SoftLink("dataset")
link_attrs["links"]["relative_softlink"] = h5py.SoftLink("group/dataset")
link_attrs["links"]["absolute_softlink"] = h5py.SoftLink("/links/group/dataset")
link_attrs["links"]["external_link"] = h5py.ExternalLink(ext_filename, "/ext_group/dataset")


class DictTestCase(unittest.TestCase):

    def assertRecursiveEqual(self, expected, actual, nodes=tuple()):
        err_msg = "\n\n Tree nodes: {}".format(nodes)
        if isinstance(expected, dict):
            self.assertTrue(isinstance(actual, dict), msg=err_msg)
            self.assertEqual(
                set(expected.keys()),
                set(actual.keys()),
                msg=err_msg
            )
            for k in actual:
                self.assertRecursiveEqual(
                    expected[k],
                    actual[k],
                    nodes=nodes + (k,),
                )
            return
        if isinstance(actual, numpy.ndarray):
            actual = actual.tolist()
        if isinstance(expected, numpy.ndarray):
            expected = expected.tolist()

        self.assertEqual(expected, actual, msg=err_msg)


class H5DictTestCase(DictTestCase):

    def _dictRoundTripNormalize(self, treedict):
        """Convert the dictionary as expected from a round-trip
        treedict -> dicttoh5 -> h5todict -> newtreedict
        """
        for key, value in list(treedict.items()):
            if isinstance(value, dict):
                self._dictRoundTripNormalize(value)

        # Expand treedict[("group", "attr_name")]
        #     to treedict["group"]["attr_name"]
        for key, value in list(treedict.items()):
            if not isinstance(key, tuple):
                continue
            # Put the attribute inside the group
            grpname, attr = key
            if not grpname:
                continue
            group = treedict.setdefault(grpname, dict())
            if isinstance(group, dict):
                del treedict[key]
                group[("", attr)] = value

    def dictRoundTripNormalize(self, treedict):
        treedict2 = deepcopy(treedict)
        self._dictRoundTripNormalize(treedict2)
        return treedict2


class TestDictToH5(H5DictTestCase):
    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        self.h5_fname = os.path.join(self.tempdir, "cityattrs.h5")
        self.h5_ext_fname = os.path.join(self.tempdir, ext_filename)

    def tearDown(self):
        if os.path.exists(self.h5_fname):
            os.unlink(self.h5_fname)
        if os.path.exists(self.h5_ext_fname):
            os.unlink(self.h5_ext_fname)
        os.rmdir(self.tempdir)

    def testH5CityAttrs(self):
        filters = {'shuffle': True,
                   'fletcher32': True}
        dicttoh5(city_attrs, self.h5_fname, h5path='/city attributes',
                 mode="w", create_dataset_args=filters)

        h5f = h5py.File(self.h5_fname, mode='r')

        self.assertIn("Tourcoing/area", h5f["/city attributes/Europe/France"])
        ds = h5f["/city attributes/Europe/France/Grenoble/inhabitants"]
        self.assertEqual(ds[...], 160215)

        # filters only apply to datasets that are not scalars (shape != () )
        ds = h5f["/city attributes/Europe/France/Grenoble/coordinates"]
        #self.assertEqual(ds.compression, "gzip")
        self.assertTrue(ds.fletcher32)
        self.assertTrue(ds.shuffle)

        h5f.close()

        ddict = load(self.h5_fname, fmat="hdf5")
        self.assertAlmostEqual(
                min(ddict["city attributes"]["Europe"]["France"]["Grenoble"]["coordinates"]),
                5.7196)

    def testH5OverwriteDeprecatedApi(self):
        dd = ConfigDict({'t': True})

        dicttoh5(h5file=self.h5_fname, treedict=dd, mode='a')
        dd = ConfigDict({'t': False})
        dicttoh5(h5file=self.h5_fname, treedict=dd, mode='a',
                 overwrite_data=False)

        res = h5todict(self.h5_fname)
        assert(res['t'] == True)

        dicttoh5(h5file=self.h5_fname, treedict=dd, mode='a',
                 overwrite_data=True)

        res = h5todict(self.h5_fname)
        assert(res['t'] == False)

    def testAttributes(self):
        """Any kind of attribute can be described"""
        ddict = {
            "group": {"datatset": "hmmm", ("", "group_attr"): 10},
            "dataset": "aaaaaaaaaaaaaaa",
            ("", "root_attr"): 11,
            ("dataset", "dataset_attr"): 12,
            ("group", "group_attr2"): 13,
        }
        with h5py.File(self.h5_fname, "w") as h5file:
            dictdump.dicttoh5(ddict, h5file)
            self.assertEqual(h5file["group"].attrs['group_attr'], 10)
            self.assertEqual(h5file.attrs['root_attr'], 11)
            self.assertEqual(h5file["dataset"].attrs['dataset_attr'], 12)
            self.assertEqual(h5file["group"].attrs['group_attr2'], 13)

    def testPathAttributes(self):
        """A group is requested at a path"""
        ddict = {
            ("", "NX_class"): 'NXcollection',
        }
        with h5py.File(self.h5_fname, "w") as h5file:
            # This should not warn
            with LoggingValidator(dictdump_logger, warning=0):
                dictdump.dicttoh5(ddict, h5file, h5path="foo/bar")

    def testKeyOrder(self):
        ddict1 = {
            "d": "plow",
            ("d", "a"): "ox",
        }
        ddict2 = {
            ("d", "a"): "ox",
            "d": "plow",
        }
        with h5py.File(self.h5_fname, "w") as h5file:
            dictdump.dicttoh5(ddict1, h5file, h5path="g1")
            dictdump.dicttoh5(ddict2, h5file, h5path="g2")
            self.assertEqual(h5file["g1/d"].attrs['a'], "ox")
            self.assertEqual(h5file["g2/d"].attrs['a'], "ox")

    def testAttributeValues(self):
        """Any NX data types can be used"""
        ddict = {
            ("", "bool"): True,
            ("", "int"): 11,
            ("", "float"): 1.1,
            ("", "str"): "a",
            ("", "boollist"): [True, False, True],
            ("", "intlist"): [11, 22, 33],
            ("", "floatlist"): [1.1, 2.2, 3.3],
            ("", "strlist"): ["a", "bb", "ccc"],
        }
        with h5py.File(self.h5_fname, "w") as h5file:
            dictdump.dicttoh5(ddict, h5file)
            for k, expected in ddict.items():
                result = h5file.attrs[k[1]]
                if isinstance(expected, list):
                    if isinstance(expected[0], str):
                        numpy.testing.assert_array_equal(result, expected)
                    else:
                        numpy.testing.assert_array_almost_equal(result, expected)
                else:
                    self.assertEqual(result, expected)

    def testAttributeAlreadyExists(self):
        """A duplicated attribute warns if overwriting is not enabled"""
        ddict = {
            "group": {"dataset": "hmmm", ("", "attr"): 10},
            ("group", "attr"): 10,
        }
        with h5py.File(self.h5_fname, "w") as h5file:
            dictdump.dicttoh5(ddict, h5file)
            self.assertEqual(h5file["group"].attrs['attr'], 10)

    def testFlatDict(self):
        """Description of a tree with a single level of keys"""
        ddict = {
            "group/group/dataset": 10,
            ("group/group/dataset", "attr"): 11,
            ("group/group", "attr"): 12,
        }
        with h5py.File(self.h5_fname, "w") as h5file:
            dictdump.dicttoh5(ddict, h5file)
            self.assertEqual(h5file["group/group/dataset"][()], 10)
            self.assertEqual(h5file["group/group/dataset"].attrs['attr'], 11)
            self.assertEqual(h5file["group/group"].attrs['attr'], 12)

    def testLinks(self):
        with h5py.File(self.h5_ext_fname, "w") as h5file:
            dictdump.dicttoh5(ext_attrs, h5file)
        with h5py.File(self.h5_fname, "w") as h5file:
            dictdump.dicttoh5(link_attrs, h5file)
        with h5py.File(self.h5_fname, "r") as h5file:
            self.assertEqual(h5file["links/group/dataset"][()], 10)
            self.assertEqual(h5file["links/group/relative_softlink"][()], 10)
            self.assertEqual(h5file["links/relative_softlink"][()], 10)
            self.assertEqual(h5file["links/absolute_softlink"][()], 10)
            self.assertEqual(h5file["links/external_link"][()], 10)

    def testDumpNumpyArray(self):
        ddict = {
            'darks': {
                '0': numpy.array([[0, 0, 0], [0, 0, 0]], dtype=numpy.uint16)
            }
        }
        with h5py.File(self.h5_fname, "w") as h5file:
            dictdump.dicttoh5(ddict, h5file)
        with h5py.File(self.h5_fname, "r") as h5file:
            numpy.testing.assert_array_equal(h5py_read_dataset(h5file["darks"]["0"]),
                                             ddict['darks']['0'])

    def testOverwrite(self):
        # Tree structure that will be tested
        group1 = {
            ("", "attr2"): "original2",
            "dset1": 0,
            "dset2": [0, 1],
            ("dset1", "attr1"): "original1",
            ("dset1", "attr2"): "original2",
            ("dset2", "attr1"): "original1",
            ("dset2", "attr2"): "original2",
        }
        group2 = {
            "subgroup1": group1.copy(),
            "subgroup2": group1.copy(),
            ("subgroup1", "attr1"): "original1",
            ("subgroup2", "attr1"): "original1"
        }
        group2.update(group1)
        # initial HDF5 tree
        otreedict = {
            ('', 'attr1'): "original1",
            ('', 'attr2'): "original2",
            'group1': group1,
            'group2': group2,
            ('group1', 'attr1'): "original1",
            ('group2', 'attr1'): "original1"
        }
        wtreedict = None  # dumped dictionary
        etreedict = None  # expected HDF5 tree after dump

        def reset_file():
            dicttoh5(
                otreedict,
                h5file=self.h5_fname,
                mode="w",
            )

        def append_file(update_mode):
            dicttoh5(
                wtreedict,
                h5file=self.h5_fname,
                mode="a",
                update_mode=update_mode
            )

        def assert_file():
            rtreedict = h5todict(
                self.h5_fname,
                include_attributes=True,
                asarray=False
            )
            netreedict = self.dictRoundTripNormalize(etreedict)
            try:
                self.assertRecursiveEqual(netreedict, rtreedict)
            except AssertionError:
                from pprint import pprint
                print("\nDUMP:")
                pprint(wtreedict)
                print("\nEXPECTED:")
                pprint(netreedict)
                print("\nHDF5:")
                pprint(rtreedict)
                raise

        def assert_append(update_mode):
            append_file(update_mode)
            assert_file()

        # Test wrong arguments
        with self.assertRaises(ValueError):
            dicttoh5(
                otreedict,
                h5file=self.h5_fname,
                mode="w",
                update_mode="wrong-value"
            )

        # No writing
        reset_file()
        etreedict = deepcopy(otreedict)
        assert_file()

        # Write identical dictionary
        wtreedict = deepcopy(otreedict)

        reset_file()
        etreedict = deepcopy(otreedict)
        for update_mode in [None, "add", "modify", "replace"]:
            assert_append(update_mode)

        # Write empty dictionary
        wtreedict = dict()

        reset_file()
        etreedict = deepcopy(otreedict)
        for update_mode in [None, "add", "modify", "replace"]:
            assert_append(update_mode)

        # Modified dataset
        wtreedict = dict()
        wtreedict["group2"] = dict()
        wtreedict["group2"]["subgroup2"] = dict()
        wtreedict["group2"]["subgroup2"]["dset1"] = {"dset3": [10, 20]}
        wtreedict["group2"]["subgroup2"]["dset2"] = [10, 20]

        reset_file()
        etreedict = deepcopy(otreedict)
        for update_mode in [None, "add"]:
            assert_append(update_mode)

        etreedict["group2"]["subgroup2"]["dset2"] = [10, 20]
        assert_append("modify")

        etreedict["group2"] = dict()
        del etreedict[("group2", "attr1")]
        etreedict["group2"]["subgroup2"] = dict()
        etreedict["group2"]["subgroup2"]["dset1"] = {"dset3": [10, 20]}
        etreedict["group2"]["subgroup2"]["dset2"] = [10, 20]
        assert_append("replace")

        # Modified group
        wtreedict = dict()
        wtreedict["group2"] = dict()
        wtreedict["group2"]["subgroup2"] = [0, 1]

        reset_file()
        etreedict = deepcopy(otreedict)
        for update_mode in [None, "add", "modify"]:
            assert_append(update_mode)

        etreedict["group2"] = dict()
        del etreedict[("group2", "attr1")]
        etreedict["group2"]["subgroup2"] = [0, 1]
        assert_append("replace")

        # Modified attribute
        wtreedict = dict()
        wtreedict["group2"] = dict()
        wtreedict["group2"]["subgroup2"] = dict()
        wtreedict["group2"]["subgroup2"][("dset1", "attr1")] = "modified"

        reset_file()
        etreedict = deepcopy(otreedict)
        for update_mode in [None, "add"]:
            assert_append(update_mode)

        etreedict["group2"]["subgroup2"][("dset1", "attr1")] = "modified"
        assert_append("modify")

        etreedict["group2"] = dict()
        del etreedict[("group2", "attr1")]
        etreedict["group2"]["subgroup2"] = dict()
        etreedict["group2"]["subgroup2"]["dset1"] = dict()
        etreedict["group2"]["subgroup2"]["dset1"][("", "attr1")] = "modified"
        assert_append("replace")

        # Delete group
        wtreedict = dict()
        wtreedict["group2"] = dict()
        wtreedict["group2"]["subgroup2"] = None

        reset_file()
        etreedict = deepcopy(otreedict)
        for update_mode in [None, "add"]:
            assert_append(update_mode)

        del etreedict["group2"]["subgroup2"]
        del etreedict["group2"][("subgroup2", "attr1")]
        assert_append("modify")

        etreedict["group2"] = dict()
        del etreedict[("group2", "attr1")]
        assert_append("replace")

        # Delete dataset
        wtreedict = dict()
        wtreedict["group2"] = dict()
        wtreedict["group2"]["subgroup2"] = dict()
        wtreedict["group2"]["subgroup2"]["dset2"] = None

        reset_file()
        etreedict = deepcopy(otreedict)
        for update_mode in [None, "add"]:
            assert_append(update_mode)

        del etreedict["group2"]["subgroup2"]["dset2"]
        del etreedict["group2"]["subgroup2"][("dset2", "attr1")]
        del etreedict["group2"]["subgroup2"][("dset2", "attr2")]
        assert_append("modify")

        etreedict["group2"] = dict()
        del etreedict[("group2", "attr1")]
        etreedict["group2"]["subgroup2"] = dict()
        assert_append("replace")

        # Delete attribute
        wtreedict = dict()
        wtreedict["group2"] = dict()
        wtreedict["group2"]["subgroup2"] = dict()
        wtreedict["group2"]["subgroup2"][("dset2", "attr1")] = None

        reset_file()
        etreedict = deepcopy(otreedict)
        for update_mode in [None, "add"]:
            assert_append(update_mode)

        del etreedict["group2"]["subgroup2"][("dset2", "attr1")]
        assert_append("modify")

        etreedict["group2"] = dict()
        del etreedict[("group2", "attr1")]
        etreedict["group2"]["subgroup2"] = dict()
        etreedict["group2"]["subgroup2"]["dset2"] = dict()
        assert_append("replace")


@pytest.mark.skipif(pint is None, reason="Require pint")
def test_dicttoh5_pint(tmp_h5py_file):
    ureg = pint.UnitRegistry()
    treedict = {
        "array_mm": pint.Quantity([1, 2, 3], ureg.mm),
        "value_kg": 3 * ureg.kg,
    }

    dicttoh5(treedict, tmp_h5py_file)

    result = h5todict(tmp_h5py_file)
    assert set(treedict.keys()) == set(result.keys())
    for key, value in treedict.items():
        assert numpy.array_equal(result[key], value.magnitude)


class TestH5ToDict(H5DictTestCase):
    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        self.h5_fname = os.path.join(self.tempdir, "cityattrs.h5")
        self.h5_ext_fname = os.path.join(self.tempdir, ext_filename)
        dicttoh5(city_attrs, self.h5_fname)
        dicttoh5(link_attrs, self.h5_fname, mode="a")
        dicttoh5(ext_attrs, self.h5_ext_fname)

    def tearDown(self):
        if os.path.exists(self.h5_fname):
            os.unlink(self.h5_fname)
        if os.path.exists(self.h5_ext_fname):
            os.unlink(self.h5_ext_fname)
        os.rmdir(self.tempdir)

    def testExcludeNames(self):
        ddict = h5todict(self.h5_fname, path="/Europe/France",
                         exclude_names=["ourcoing", "inhab", "toto"])
        self.assertNotIn("Tourcoing", ddict)
        self.assertIn("Grenoble", ddict)

        self.assertNotIn("inhabitants", ddict["Grenoble"])
        self.assertIn("coordinates", ddict["Grenoble"])
        self.assertIn("area", ddict["Grenoble"])

    def testAsArrayTrue(self):
        """Test with asarray=True, the default"""
        ddict = h5todict(self.h5_fname, path="/Europe/France/Grenoble")
        self.assertTrue(numpy.array_equal(ddict["inhabitants"], numpy.array(inhabitants)))

    def testAsArrayFalse(self):
        """Test with asarray=False"""
        ddict = h5todict(self.h5_fname, path="/Europe/France/Grenoble", asarray=False)
        self.assertEqual(ddict["inhabitants"], inhabitants)

    def testDereferenceLinks(self):
        ddict = h5todict(self.h5_fname, path="links", dereference_links=True)
        self.assertTrue(ddict["absolute_softlink"], 10)
        self.assertTrue(ddict["relative_softlink"], 10)
        self.assertTrue(ddict["external_link"], 10)
        self.assertTrue(ddict["group"]["relative_softlink"], 10)

    def testPreserveLinks(self):
        ddict = h5todict(self.h5_fname, path="links", dereference_links=False)
        self.assertTrue(is_link(ddict["absolute_softlink"]))
        self.assertTrue(is_link(ddict["relative_softlink"]))
        self.assertTrue(is_link(ddict["external_link"]))
        self.assertTrue(is_link(ddict["group"]["relative_softlink"]))

    def testStrings(self):
        ddict = {"dset_bytes": b"bytes",
                 "dset_utf8": "utf8",
                 "dset_2bytes": [b"bytes", b"bytes"],
                 "dset_2utf8": ["utf8", "utf8"],
                 ("", "attr_bytes"): b"bytes",
                 ("", "attr_utf8"): "utf8",
                 ("", "attr_2bytes"): [b"bytes", b"bytes"],
                 ("", "attr_2utf8"): ["utf8", "utf8"]}
        dicttoh5(ddict, self.h5_fname, mode="w")
        adict = h5todict(self.h5_fname, include_attributes=True, asarray=False)
        self.assertEqual(ddict["dset_bytes"], adict["dset_bytes"])
        self.assertEqual(ddict["dset_utf8"], adict["dset_utf8"])
        self.assertEqual(ddict[("", "attr_bytes")], adict[("", "attr_bytes")])
        self.assertEqual(ddict[("", "attr_utf8")], adict[("", "attr_utf8")])
        numpy.testing.assert_array_equal(ddict["dset_2bytes"], adict["dset_2bytes"])
        numpy.testing.assert_array_equal(ddict["dset_2utf8"], adict["dset_2utf8"])
        numpy.testing.assert_array_equal(ddict[("", "attr_2bytes")], adict[("", "attr_2bytes")])
        numpy.testing.assert_array_equal(ddict[("", "attr_2utf8")], adict[("", "attr_2utf8")])


class TestDictToNx(H5DictTestCase):
    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        self.h5_fname = os.path.join(self.tempdir, "nx.h5")
        self.h5_ext_fname = os.path.join(self.tempdir, "nx_ext.h5")

    def tearDown(self):
        if os.path.exists(self.h5_fname):
            os.unlink(self.h5_fname)
        if os.path.exists(self.h5_ext_fname):
            os.unlink(self.h5_ext_fname)
        os.rmdir(self.tempdir)

    def testAttributes(self):
        """Any kind of attribute can be described"""
        ddict = {
            "group": {"dataset": 100, "@group_attr1": 10},
            "dataset": 200,
            "@root_attr": 11,
            "dataset@dataset_attr": "12",
            "group@group_attr2": 13,
        }
        with h5py.File(self.h5_fname, "w") as h5file:
            dictdump.dicttonx(ddict, h5file)
            self.assertEqual(h5file["group"].attrs['group_attr1'], 10)
            self.assertEqual(h5file.attrs['root_attr'], 11)
            self.assertEqual(h5file["dataset"].attrs['dataset_attr'], "12")
            self.assertEqual(h5file["group"].attrs['group_attr2'], 13)

    def testKeyOrder(self):
        ddict1 = {
            "d": "plow",
            "d@a": "ox",
        }
        ddict2 = {
            "d@a": "ox",
            "d": "plow",
        }
        with h5py.File(self.h5_fname, "w") as h5file:
            dictdump.dicttonx(ddict1, h5file, h5path="g1")
            dictdump.dicttonx(ddict2, h5file, h5path="g2")
            self.assertEqual(h5file["g1/d"].attrs['a'], "ox")
            self.assertEqual(h5file["g2/d"].attrs['a'], "ox")

    def testAttributeValues(self):
        """Any NX data types can be used"""
        ddict = {
            "@bool": True,
            "@int": 11,
            "@float": 1.1,
            "@str": "a",
            "@boollist": [True, False, True],
            "@intlist": [11, 22, 33],
            "@floatlist": [1.1, 2.2, 3.3],
            "@strlist": ["a", "bb", "ccc"],
        }
        with h5py.File(self.h5_fname, "w") as h5file:
            dictdump.dicttonx(ddict, h5file)
            for k, expected in ddict.items():
                result = h5file.attrs[k[1:]]
                if isinstance(expected, list):
                    if isinstance(expected[0], str):
                        numpy.testing.assert_array_equal(result, expected)
                    else:
                        numpy.testing.assert_array_almost_equal(result, expected)
                else:
                    self.assertEqual(result, expected)

    def testFlatDict(self):
        """Description of a tree with a single level of keys"""
        ddict = {
            "group/group/dataset": 10,
            "group/group/dataset@attr": 11,
            "group/group@attr": 12,
        }
        with h5py.File(self.h5_fname, "w") as h5file:
            dictdump.dicttonx(ddict, h5file)
            self.assertEqual(h5file["group/group/dataset"][()], 10)
            self.assertEqual(h5file["group/group/dataset"].attrs['attr'], 11)
            self.assertEqual(h5file["group/group"].attrs['attr'], 12)

    def testLinks(self):
        ddict = {"ext_group": {"dataset": 10}}
        dictdump.dicttonx(ddict, self.h5_ext_fname)
        ddict = {"links": {"group": {"dataset": 10, ">relative_softlink": "dataset"},
                           ">relative_softlink": "group/dataset",
                           ">absolute_softlink": "/links/group/dataset",
                           ">external_link": "nx_ext.h5::/ext_group/dataset"}}
        dictdump.dicttonx(ddict, self.h5_fname)
        with h5py.File(self.h5_fname, "r") as h5file:
            self.assertEqual(h5file["links/group/dataset"][()], 10)
            self.assertEqual(h5file["links/group/relative_softlink"][()], 10)
            self.assertEqual(h5file["links/relative_softlink"][()], 10)
            self.assertEqual(h5file["links/absolute_softlink"][()], 10)
            self.assertEqual(h5file["links/external_link"][()], 10)

    def testUpLinks(self):
        ddict = {"data": {"group": {"dataset": 10, ">relative_softlink": "dataset"}},
                 "links": {"group": {"subgroup": {">relative_softlink": "../../../data/group/dataset"}}}}
        dictdump.dicttonx(ddict, self.h5_fname)
        with h5py.File(self.h5_fname, "r") as h5file:
            self.assertEqual(h5file["/links/group/subgroup/relative_softlink"][()], 10)

    def testOverwrite(self):
        entry_name = "entry"
        wtreedict = {
            "group1": {"a": 1, "b": 2},
            "group2@attr3": "attr3",
            "group2@attr4": "attr4",
            "group2": {
                "@attr1": "attr1",
                "@attr2": "attr2",
                "c": 3,
                "d": 4,
                "dataset4": 8,
                "dataset4@units": "keV",
            },
            "group3": {"subgroup": {"e": 9, "f": 10}},
            "dataset1": 5,
            "dataset2": 6,
            "dataset3": 7,
            "dataset3@units": "mm",
        }
        esubtree = {
            "@NX_class": "NXentry",
            "group1": {"@NX_class": "NXcollection", "a": 1, "b": 2},
            "group2": {
                "@NX_class": "NXcollection",
                "@attr1": "attr1",
                "@attr2": "attr2",
                "@attr3": "attr3",
                "@attr4": "attr4",
                "c": 3,
                "d": 4,
                "dataset4": 8,
                "dataset4@units": "keV",
            },
            "group3": {
                "@NX_class": "NXcollection",
                "subgroup": {"@NX_class": "NXcollection", "e": 9, "f": 10},
            },
            "dataset1": 5,
            "dataset2": 6,
            "dataset3": 7,
            "dataset3@units": "mm",
        }
        etreedict = {entry_name: esubtree}

        def append_file(update_mode, add_nx_class):
            dictdump.dicttonx(
                wtreedict,
                h5file=self.h5_fname,
                mode="a",
                h5path=entry_name,
                update_mode=update_mode,
                add_nx_class=add_nx_class
            )

        def assert_file():
            rtreedict = dictdump.nxtodict(
                self.h5_fname,
                include_attributes=True,
                asarray=False,
            )
            netreedict = self.dictRoundTripNormalize(etreedict)
            try:
                self.assertRecursiveEqual(netreedict, rtreedict)
            except AssertionError:
                from pprint import pprint
                print("\nDUMP:")
                pprint(wtreedict)
                print("\nEXPECTED:")
                pprint(netreedict)
                print("\nHDF5:")
                pprint(rtreedict)
                raise

        def assert_append(update_mode, add_nx_class=None):
            append_file(update_mode, add_nx_class=add_nx_class)
            assert_file()

        # First to an empty file
        assert_append(None)

        # Add non-existing attributes/datasets/groups
        wtreedict["group1"].pop("a")
        wtreedict["group2"].pop("@attr1")
        wtreedict["group2"]["@attr2"] = "attr3"  # only for update
        wtreedict["group2"]["@type"] = "test"
        wtreedict["group2"]["dataset4"] = 9  # only for update
        del wtreedict["group2"]["dataset4@units"]
        wtreedict["group3"] = {}
        esubtree["group2"]["@type"] = "test"
        assert_append("add")

        # Add update existing attributes and datasets
        esubtree["group2"]["@attr2"] = "attr3"
        esubtree["group2"]["dataset4"] = 9
        assert_append("modify")

        # Do not add missing NX_class by default when updating
        wtreedict["group2"]["@NX_class"] = "NXprocess"
        esubtree["group2"]["@NX_class"] = "NXprocess"
        assert_append("modify")
        del wtreedict["group2"]["@NX_class"]
        assert_append("modify")

        # Overwrite existing groups/datasets/attributes
        esubtree["group1"].pop("a")
        esubtree["group2"].pop("@attr1")
        esubtree["group2"]["@NX_class"] = "NXcollection"
        esubtree["group2"]["dataset4"] = 9
        del esubtree["group2"]["dataset4@units"]
        esubtree["group3"] = {"@NX_class": "NXcollection"}
        assert_append("replace", add_nx_class=True)


@pytest.mark.skipif(pint is None, reason="Require pint")
def test_dicttonx_pint(tmp_h5py_file):
    ureg = pint.UnitRegistry()
    treedict = {
        "array_mm": pint.Quantity([1, 2, 3], ureg.mm),
        "value_kg": 3 * ureg.kg,
    }

    dictdump.dicttonx(treedict, tmp_h5py_file)

    result = dictdump.nxtodict(tmp_h5py_file)
    for key, value in treedict.items():
        assert numpy.array_equal(result[key], value.magnitude)
        assert result[f"{key}@units"] == f"{value.units:~C}"


class TestNxToDict(H5DictTestCase):
    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        self.h5_fname = os.path.join(self.tempdir, "nx.h5")
        self.h5_ext_fname = os.path.join(self.tempdir, "nx_ext.h5")

    def tearDown(self):
        if os.path.exists(self.h5_fname):
            os.unlink(self.h5_fname)
        if os.path.exists(self.h5_ext_fname):
            os.unlink(self.h5_ext_fname)
        os.rmdir(self.tempdir)

    def testAttributes(self):
        """Any kind of attribute can be described"""
        ddict = {
            "group": {"dataset": 100, "@group_attr1": 10},
            "dataset": 200,
            "@root_attr": 11,
            "dataset@dataset_attr": "12",
            "group@group_attr2": 13,
        }
        dictdump.dicttonx(ddict, self.h5_fname)
        ddict = dictdump.nxtodict(self.h5_fname, include_attributes=True)
        self.assertEqual(ddict["group"]["@group_attr1"], 10)
        self.assertEqual(ddict["@root_attr"], 11)
        self.assertEqual(ddict["dataset@dataset_attr"], "12")
        self.assertEqual(ddict["group"]["@group_attr2"], 13)

    def testDereferenceLinks(self):
        """Write links and dereference on read"""
        ddict = {"ext_group": {"dataset": 10}}
        dictdump.dicttonx(ddict, self.h5_ext_fname)
        ddict = {"links": {"group": {"dataset": 10, ">relative_softlink": "dataset"},
                           ">relative_softlink": "group/dataset",
                           ">absolute_softlink": "/links/group/dataset",
                           ">external_link": "nx_ext.h5::/ext_group/dataset"}}
        dictdump.dicttonx(ddict, self.h5_fname)

        ddict = dictdump.h5todict(self.h5_fname, dereference_links=True)
        self.assertTrue(ddict["links"]["absolute_softlink"], 10)
        self.assertTrue(ddict["links"]["relative_softlink"], 10)
        self.assertTrue(ddict["links"]["external_link"], 10)
        self.assertTrue(ddict["links"]["group"]["relative_softlink"], 10)

    def testPreserveLinks(self):
        """Write/read links"""
        ddict = {"ext_group": {"dataset": 10}}
        dictdump.dicttonx(ddict, self.h5_ext_fname)
        ddict = {"links": {"group": {"dataset": 10, ">relative_softlink": "dataset"},
                           ">relative_softlink": "group/dataset",
                           ">absolute_softlink": "/links/group/dataset",
                           ">external_link": "nx_ext.h5::/ext_group/dataset"}}
        dictdump.dicttonx(ddict, self.h5_fname)

        ddict = dictdump.nxtodict(self.h5_fname, dereference_links=False)
        self.assertTrue(ddict["links"][">absolute_softlink"], "dataset")
        self.assertTrue(ddict["links"][">relative_softlink"], "group/dataset")
        self.assertTrue(ddict["links"][">external_link"], "/links/group/dataset")
        self.assertTrue(ddict["links"]["group"][">relative_softlink"], "nx_ext.h5::/ext_group/datase")

    def testNotExistingPath(self):
        """Test converting not existing path"""
        with h5py.File(self.h5_fname, 'a') as f:
            f['data'] = 1

        ddict = h5todict(self.h5_fname, path="/I/am/not/a/path", errors='ignore')
        self.assertFalse(ddict)

        with LoggingValidator(dictdump_logger, error=1):
            ddict = h5todict(self.h5_fname, path="/I/am/not/a/path", errors='log')
            self.assertFalse(ddict)

        with self.assertRaises(KeyError):
            h5todict(self.h5_fname, path="/I/am/not/a/path", errors='raise')

    def testBrokenLinks(self):
        """Test with broken links"""
        with h5py.File(self.h5_fname, 'a') as f:
            f["/Mars/BrokenSoftLink"] = h5py.SoftLink("/Idontexists")
            f["/Mars/BrokenExternalLink"] = h5py.ExternalLink("notexistingfile.h5", "/Idontexists")

        ddict = h5todict(self.h5_fname, path="/Mars", errors='ignore')
        self.assertFalse(ddict)

        with LoggingValidator(dictdump_logger, error=2):
            ddict = h5todict(self.h5_fname, path="/Mars", errors='log')
            self.assertFalse(ddict)

        with self.assertRaises(KeyError):
            h5todict(self.h5_fname, path="/Mars", errors='raise')


class TestDictToJson(DictTestCase):
    def setUp(self):
        self.dir_path = tempfile.mkdtemp()
        self.json_fname = os.path.join(self.dir_path, "cityattrs.json")

    def tearDown(self):
        os.unlink(self.json_fname)
        os.rmdir(self.dir_path)

    def testJsonCityAttrs(self):
        self.json_fname = os.path.join(self.dir_path, "cityattrs.json")
        dicttojson(city_attrs, self.json_fname, indent=3)

        with open(self.json_fname, "r") as f:
            json_content = f.read()
            self.assertIn('"inhabitants": 160215', json_content)


class TestDictToIni(DictTestCase):
    def setUp(self):
        self.dir_path = tempfile.mkdtemp()
        self.ini_fname = os.path.join(self.dir_path, "test.ini")

    def tearDown(self):
        os.unlink(self.ini_fname)
        os.rmdir(self.dir_path)

    def testConfigDictIO(self):
        """Ensure values and types of data is preserved when dictionary is
        written to file and read back."""
        testdict = {
            'simple_types': {
                'float': 1.0,
                'int': 1,
                'percent string': '5 % is too much',
                'backslash string': 'i can use \\',
                'empty_string': '',
                'nonestring': 'None',
                'nonetype': None,
                'interpstring': 'interpolation: %(percent string)s',
            },
            'containers': {
                'list': [-1, 'string', 3.0, False, None],
                'array': numpy.array([1.0, 2.0, 3.0]),
                'dict': {
                    'key1': 'Hello World',
                    'key2': 2.0,
                }
            }
        }

        dump(testdict, self.ini_fname)

        #read the data back
        readdict = load(self.ini_fname)

        testdictkeys = list(testdict.keys())
        readkeys = list(readdict.keys())

        self.assertTrue(len(readkeys) == len(testdictkeys),
                        "Number of read keys not equal")

        self.assertEqual(readdict['simple_types']["interpstring"],
                         "interpolation: 5 % is too much")

        testdict['simple_types']["interpstring"] = "interpolation: 5 % is too much"

        for key in testdict["simple_types"]:
            original = testdict['simple_types'][key]
            read = readdict['simple_types'][key]
            self.assertEqual(read, original,
                             "Read <%s> instead of <%s>" % (read, original))

        for key in testdict["containers"]:
            original = testdict["containers"][key]
            read = readdict["containers"][key]
            if key == 'array':
                self.assertEqual(read.all(), original.all(),
                            "Read <%s> instead of <%s>" % (read, original))
            else:
                self.assertEqual(read, original,
                            "Read <%s> instead of <%s>" % (read, original))

    def testConfigDictOrder(self):
        """Ensure order is preserved when dictionary is
        written to file and read back."""
        test_dict = {'banana': 3, 'apple': 4, 'pear': 1, 'orange': 2}
        # sort by key
        test_ordered_dict1 = OrderedDict(sorted(test_dict.items(),
                                                key=lambda t: t[0]))
        # sort by value
        test_ordered_dict2 = OrderedDict(sorted(test_dict.items(),
                                                key=lambda t: t[1]))
        # add the two ordered dict as sections of a third ordered dict
        test_ordered_dict3 = OrderedDict()
        test_ordered_dict3["section1"] = test_ordered_dict1
        test_ordered_dict3["section2"] = test_ordered_dict2

        # write to ini and read back as a ConfigDict (inherits OrderedDict)
        dump(test_ordered_dict3,
             self.ini_fname, fmat="ini")
        read_instance = ConfigDict()
        read_instance.read(self.ini_fname)

        # loop through original and read-back dictionaries,
        # test identical order for key/value pairs
        for orig_key, section in zip(test_ordered_dict3.keys(),
                                     read_instance.keys()):
            self.assertEqual(orig_key, section)
            for orig_key2, read_key in zip(test_ordered_dict3[section].keys(),
                                           read_instance[section].keys()):
                self.assertEqual(orig_key2, read_key)
                self.assertEqual(test_ordered_dict3[section][orig_key2],
                                 read_instance[section][read_key])



# /*##########################################################################
#
# Copyright (c) 2004-2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Basic tests for PlotWidget"""

__authors__ = ["T. Rueter", "T. Vincent"]
__license__ = "MIT"
__date__ = "15/05/2017"


import logging
import unittest

from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt
from silx.gui.plot import LegendSelector


_logger = logging.getLogger(__name__)


class TestLegendSelector(TestCaseQt):
    """Basic test for LegendSelector"""

    def testLegendSelector(self):
        """Test copied from __main__ of LegendSelector in PyMca"""
        class Notifier(qt.QObject):
            def __init__(self):
                qt.QObject.__init__(self)
                self.chk = True

            def signalReceived(self, **kw):
                obj = self.sender()
                _logger.info('NOTIFIER -- signal received\n\tsender: %s',
                             str(obj))

        notifier = Notifier()

        legends = ['Legend0',
                   'Legend1',
                   'Long Legend 2',
                   'Foo Legend 3',
                   'Even Longer Legend 4',
                   'Short Leg 5',
                   'Dot symbol 6',
                   'Comma symbol 7']
        colors = [qt.Qt.darkRed, qt.Qt.green, qt.Qt.yellow, qt.Qt.darkCyan,
                  qt.Qt.blue, qt.Qt.darkBlue, qt.Qt.red, qt.Qt.darkYellow]
        symbols = ['o', 't', '+', 'x', 's', 'd', '.', ',']

        win = LegendSelector.LegendListView()
        # win = LegendListContextMenu()
        # win = qt.QWidget()
        # layout = qt.QVBoxLayout()
        # layout.setContentsMargins(0,0,0,0)
        llist = []

        for _idx, (l, c, s) in enumerate(zip(legends, colors, symbols)):
            ddict = {
                'color': qt.QColor(c),
                'linewidth': 4,
                'symbol': s,
            }
            legend = l
            llist.append((legend, ddict))
            # item = qt.QListWidgetItem(win)
            # legendWidget = LegendListItemWidget(l)
            # legendWidget.icon.setSymbol(s)
            # legendWidget.icon.setColor(qt.QColor(c))
            # layout.addWidget(legendWidget)
            # win.setItemWidget(item, legendWidget)

        # win = LegendListItemWidget('Some Legend 1')
        # print(llist)
        model = LegendSelector.LegendModel(legendList=llist)
        win.setModel(model)
        win.setSelectionModel(qt.QItemSelectionModel(model))
        win.setContextMenu()
        # print('Edit triggers: %d'%win.editTriggers())

        # win = LegendListWidget(None, legends)
        # win[0].updateItem(ddict)
        # win.setLayout(layout)
        win.sigLegendSignal.connect(notifier.signalReceived)
        win.show()

        win.clear()
        win.setLegendList(llist)

        self.qWaitForWindowExposed(win)


class TestRenameCurveDialog(TestCaseQt):
    """Basic test for RenameCurveDialog"""

    def testDialog(self):
        """Create dialog, change name and press OK"""
        self.dialog = LegendSelector.RenameCurveDialog(
            None, 'curve1', ['curve1', 'curve2', 'curve3'])
        self.dialog.open()
        self.qWaitForWindowExposed(self.dialog)
        self.keyClicks(self.dialog.lineEdit, 'changed')
        self.mouseClick(self.dialog.okButton, qt.Qt.LeftButton)
        self.qapp.processEvents()
        ret = self.dialog.result()
        self.assertEqual(ret, qt.QDialog.Accepted)
        newName = self.dialog.getText()
        self.assertEqual(newName, 'curve1changed')
        del self.dialog



# /*##########################################################################
# Copyright (C) 2004-2021 V.A. Sole, European Synchrotron Radiation Facility
#
# This file is part of the PyMca X-ray Fluorescence Toolkit developed at
# the ESRF by the Software group.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ######################################################################### */
"""This module defines widgets used to build a fit configuration dialog.
The resulting dialog widget outputs a dictionary of configuration parameters.
"""
from silx.gui import qt

__authors__ = ["P. Knobel"]
__license__ = "MIT"
__date__ = "30/11/2016"


class TabsDialog(qt.QDialog):
    """Dialog widget containing a QTabWidget :attr:`tabWidget`
    and a buttons:

        # - buttonHelp
        - buttonDefaults
        - buttonOk
        - buttonCancel

    This dialog defines a __len__ returning the number of tabs,
    and an __iter__ method yielding the tab widgets.
    """
    def __init__(self, parent=None):
        qt.QDialog.__init__(self, parent)
        self.tabWidget = qt.QTabWidget(self)

        layout = qt.QVBoxLayout(self)
        layout.addWidget(self.tabWidget)

        layout2 = qt.QHBoxLayout(None)

        # self.buttonHelp = qt.QPushButton(self)
        # self.buttonHelp.setText("Help")
        # layout2.addWidget(self.buttonHelp)

        self.buttonDefault = qt.QPushButton(self)
        self.buttonDefault.setText("Undo changes")
        layout2.addWidget(self.buttonDefault)

        spacer = qt.QSpacerItem(20, 20,
                                qt.QSizePolicy.Expanding,
                                qt.QSizePolicy.Minimum)
        layout2.addItem(spacer)

        self.buttonOk = qt.QPushButton(self)
        self.buttonOk.setText("OK")
        layout2.addWidget(self.buttonOk)

        self.buttonCancel = qt.QPushButton(self)
        self.buttonCancel.setText("Cancel")
        layout2.addWidget(self.buttonCancel)

        layout.addLayout(layout2)

        self.buttonOk.clicked.connect(self.accept)
        self.buttonCancel.clicked.connect(self.reject)

    def __len__(self):
        """Return number of tabs"""
        return self.tabWidget.count()

    def __iter__(self):
        """Return the next tab widget in  :attr:`tabWidget` every
        time this method is called.

        :return: Tab widget
        :rtype: QWidget
        """
        for widget_index in range(len(self)):
            yield self.tabWidget.widget(widget_index)

    def addTab(self, page, label):
        """Add a new tab

        :param page: Content of new page. Must be a widget with
            a get() method returning a dictionary.
        :param str label: Tab label
        """
        self.tabWidget.addTab(page, label)

    def getTabLabels(self):
        """
        Return a list of all tab labels in :attr:`tabWidget`
        """
        return [self.tabWidget.tabText(i) for i in range(len(self))]


class TabsDialogData(TabsDialog):
    """This dialog adds a data attribute to :class:`TabsDialog`.

    Data input in widgets, such as text entries or checkboxes, is stored in an
    attribute :attr:`output` when the user clicks the OK button.

    A default dictionary can be supplied when this dialog is initialized, to
    be used as default data for :attr:`output`.
    """
    def __init__(self, parent=None, modal=True, default=None):
        """

        :param parent: Parent :class:`QWidget`
        :param modal: If `True`, dialog is modal, meaning this dialog remains
            in front of it's parent window and disables it until the user is
            done interacting with the dialog
        :param default: Default dictionary, used to initialize and reset
            :attr:`output`.
        """
        TabsDialog.__init__(self, parent)
        self.setModal(modal)
        self.setWindowTitle("Fit configuration")

        self.output = {}

        self.default = {} if default is None else default

        self.buttonDefault.clicked.connect(self._resetDefault)
        # self.keyPressEvent(qt.Qt.Key_Enter).

    def keyPressEvent(self, event):
        """Redefining this method to ignore Enter key
        (for some reason it activates buttonDefault callback which
        resets all widgets)
        """
        if event.key() in [qt.Qt.Key_Enter, qt.Qt.Key_Return]:
            return
        TabsDialog.keyPressEvent(self, event)

    def accept(self):
        """When *OK* is clicked, update :attr:`output` with data from
        various widgets
        """
        self.output.update(self.default)

        # loop over all tab widgets (uses TabsDialog.__iter__)
        for tabWidget in self:
            self.output.update(tabWidget.get())

        # avoid pathological None cases
        for key in self.output.keys():
            if self.output[key] is None:
                if key in self.default:
                    self.output[key] = self.default[key]
        super(TabsDialogData, self).accept()

    def reject(self):
        """When the *Cancel* button is clicked, reinitialize :attr:`output`
        and quit
        """
        self.setDefault()
        super(TabsDialogData, self).reject()

    def _resetDefault(self, checked):
        self.setDefault()

    def setDefault(self, newdefault=None):
        """Reinitialize :attr:`output` with :attr:`default` or with
        new dictionary ``newdefault`` if provided.
        Call :meth:`setDefault` for each tab widget, if available.
        """
        self.output = {}
        if newdefault is None:
            newdefault = self.default
        else:
            self.default = newdefault
        self.output.update(newdefault)

        for tabWidget in self:
            if hasattr(tabWidget, "setDefault"):
                tabWidget.setDefault(self.output)


class ConstraintsPage(qt.QGroupBox):
    """Checkable QGroupBox widget filled with QCheckBox widgets,
    to configure the fit estimation for standard fit theories.
    """
    def __init__(self, parent=None, title="Set constraints"):
        super(ConstraintsPage, self).__init__(parent)
        self.setTitle(title)
        self.setToolTip("Disable 'Set constraints' to remove all " +
                        "constraints on all fit parameters")
        self.setCheckable(True)

        layout = qt.QVBoxLayout(self)
        self.setLayout(layout)

        self.positiveHeightCB = qt.QCheckBox("Force positive height/area", self)
        self.positiveHeightCB.setToolTip("Fit must find positive peaks")
        layout.addWidget(self.positiveHeightCB)

        self.positionInIntervalCB = qt.QCheckBox("Force position in interval", self)
        self.positionInIntervalCB.setToolTip(
                "Fit must position peak within X limits")
        layout.addWidget(self.positionInIntervalCB)

        self.positiveFwhmCB = qt.QCheckBox("Force positive FWHM", self)
        self.positiveFwhmCB.setToolTip("Fit must find a positive FWHM")
        layout.addWidget(self.positiveFwhmCB)

        self.sameFwhmCB = qt.QCheckBox("Force same FWHM for all peaks", self)
        self.sameFwhmCB.setToolTip("Fit must find same FWHM for all peaks")
        layout.addWidget(self.sameFwhmCB)

        self.quotedEtaCB = qt.QCheckBox("Force Eta between 0 and 1", self)
        self.quotedEtaCB.setToolTip(
                "Fit must find Eta between 0 and 1 for pseudo-Voigt function")
        layout.addWidget(self.quotedEtaCB)

        layout.addStretch()

        self.setDefault()

    def setDefault(self, default_dict=None):
        """Set default state for all widgets.

        :param default_dict: If a default config dictionary is provided as
            a parameter, its values are used as default state."""
        if default_dict is None:
            default_dict = {}
        # this one uses reverse logic: if checked, NoConstraintsFlag must be False
        self.setChecked(
                not default_dict.get('NoConstraintsFlag', False))
        self.positiveHeightCB.setChecked(
                default_dict.get('PositiveHeightAreaFlag', True))
        self.positionInIntervalCB.setChecked(
                default_dict.get('QuotedPositionFlag', False))
        self.positiveFwhmCB.setChecked(
                default_dict.get('PositiveFwhmFlag', True))
        self.sameFwhmCB.setChecked(
                default_dict.get('SameFwhmFlag', False))
        self.quotedEtaCB.setChecked(
                default_dict.get('QuotedEtaFlag', False))

    def get(self):
        """Return a dictionary of constraint flags, to be processed by the
        :meth:`configure` method of the selected fit theory."""
        ddict = {
            'NoConstraintsFlag': not self.isChecked(),
            'PositiveHeightAreaFlag': self.positiveHeightCB.isChecked(),
            'QuotedPositionFlag': self.positionInIntervalCB.isChecked(),
            'PositiveFwhmFlag': self.positiveFwhmCB.isChecked(),
            'SameFwhmFlag': self.sameFwhmCB.isChecked(),
            'QuotedEtaFlag': self.quotedEtaCB.isChecked(),
        }
        return ddict


class SearchPage(qt.QWidget):
    def __init__(self, parent=None):
        super(SearchPage, self).__init__(parent)
        layout = qt.QVBoxLayout(self)

        self.manualFwhmGB = qt.QGroupBox("Define FWHM manually", self)
        self.manualFwhmGB.setCheckable(True)
        self.manualFwhmGB.setToolTip(
            "If disabled, the FWHM parameter used for peak search is " +
            "estimated based on the highest peak in the data")
        layout.addWidget(self.manualFwhmGB)
        # ------------ GroupBox fwhm--------------------------
        layout2 = qt.QHBoxLayout(self.manualFwhmGB)
        self.manualFwhmGB.setLayout(layout2)

        label = qt.QLabel("Fwhm Points", self.manualFwhmGB)
        layout2.addWidget(label)

        self.fwhmPointsSpin = qt.QSpinBox(self.manualFwhmGB)
        self.fwhmPointsSpin.setRange(0, 999999)
        self.fwhmPointsSpin.setToolTip("Typical peak fwhm (number of data points)")
        layout2.addWidget(self.fwhmPointsSpin)
        # ----------------------------------------------------

        self.manualScalingGB = qt.QGroupBox("Define scaling manually", self)
        self.manualScalingGB.setCheckable(True)
        self.manualScalingGB.setToolTip(
            "If disabled, the Y scaling used for peak search is " +
            "estimated automatically")
        layout.addWidget(self.manualScalingGB)
        # ------------ GroupBox scaling-----------------------
        layout3 = qt.QHBoxLayout(self.manualScalingGB)
        self.manualScalingGB.setLayout(layout3)

        label = qt.QLabel("Y Scaling", self.manualScalingGB)
        layout3.addWidget(label)

        self.yScalingEntry = qt.QLineEdit(self.manualScalingGB)
        self.yScalingEntry.setToolTip(
                "Data values will be multiplied by this value prior to peak" +
                " search")
        self.yScalingEntry.setValidator(qt.QDoubleValidator(self))
        layout3.addWidget(self.yScalingEntry)
        # ----------------------------------------------------

        # ------------------- grid layout --------------------
        containerWidget = qt.QWidget(self)
        layout4 = qt.QHBoxLayout(containerWidget)
        containerWidget.setLayout(layout4)

        label = qt.QLabel("Sensitivity", containerWidget)
        layout4.addWidget(label)

        self.sensitivityEntry = qt.QLineEdit(containerWidget)
        self.sensitivityEntry.setToolTip(
            "Peak search sensitivity threshold, expressed as a multiple " +
            "of the standard deviation of the noise.\nMinimum value is 1 " +
            "(to be detected, peak must be higher than the estimated noise)")
        sensivalidator = qt.QDoubleValidator(self)
        sensivalidator.setBottom(1.0)
        self.sensitivityEntry.setValidator(sensivalidator)
        layout4.addWidget(self.sensitivityEntry)
        # ----------------------------------------------------
        layout.addWidget(containerWidget)

        self.forcePeakPresenceCB = qt.QCheckBox("Force peak presence", self)
        self.forcePeakPresenceCB.setToolTip(
                "If peak search algorithm is unsuccessful, place one peak " +
                "at the maximum of the curve")
        layout.addWidget(self.forcePeakPresenceCB)

        layout.addStretch()

        self.setDefault()

    def setDefault(self, default_dict=None):
        """Set default values for all widgets.

        :param default_dict: If a default config dictionary is provided as
            a parameter, its values are used as default values."""
        if default_dict is None:
            default_dict = {}
        self.manualFwhmGB.setChecked(
                not default_dict.get('AutoFwhm', True))
        self.fwhmPointsSpin.setValue(
               default_dict.get('FwhmPoints', 8))
        self.sensitivityEntry.setText(
                str(default_dict.get('Sensitivity', 1.0)))
        self.manualScalingGB.setChecked(
                not default_dict.get('AutoScaling', False))
        self.yScalingEntry.setText(
                str(default_dict.get('Yscaling', 1.0)))
        self.forcePeakPresenceCB.setChecked(
                default_dict.get('ForcePeakPresence', False))

    def get(self):
        """Return a dictionary of peak search parameters, to be processed by
        the :meth:`configure` method of the selected fit theory."""
        ddict = {
            'AutoFwhm': not self.manualFwhmGB.isChecked(),
            'FwhmPoints': self.fwhmPointsSpin.value(),
            'Sensitivity': safe_float(self.sensitivityEntry.text()),
            'AutoScaling': not self.manualScalingGB.isChecked(),
            'Yscaling': safe_float(self.yScalingEntry.text()),
            'ForcePeakPresence': self.forcePeakPresenceCB.isChecked()
        }
        return ddict


class BackgroundPage(qt.QGroupBox):
    """Background subtraction configuration, specific to fittheories
    estimation functions."""
    def __init__(self, parent=None,
                 title="Subtract strip background prior to estimation"):
        super(BackgroundPage, self).__init__(parent)
        self.setTitle(title)
        self.setCheckable(True)
        self.setToolTip(
            "The strip algorithm strips away peaks to compute the " +
            "background signal.\nAt each iteration, a sample is compared " +
            "to the average of the two samples at a given distance in both" +
            " directions,\n and if its value is higher than the average,"
            "it is replaced by the average.")

        layout = qt.QGridLayout(self)
        self.setLayout(layout)

        for i, label_text in enumerate(
                ["Strip width (in samples)",
                 "Number of iterations",
                 "Strip threshold factor"]):
            label = qt.QLabel(label_text)
            layout.addWidget(label, i, 0)

        self.stripWidthSpin = qt.QSpinBox(self)
        self.stripWidthSpin.setToolTip(
            "Width, in number of samples, of the strip operator")
        self.stripWidthSpin.setRange(1, 999999)

        layout.addWidget(self.stripWidthSpin, 0, 1)

        self.numIterationsSpin = qt.QSpinBox(self)
        self.numIterationsSpin.setToolTip(
            "Number of iterations of the strip algorithm")
        self.numIterationsSpin.setRange(1, 999999)
        layout.addWidget(self.numIterationsSpin, 1, 1)

        self.thresholdFactorEntry = qt.QLineEdit(self)
        self.thresholdFactorEntry.setToolTip(
            "Factor used by the strip algorithm to decide whether a sample" +
            "value should be stripped.\nThe value must be higher than the " +
            "average of the 2 samples at +- w times this factor.\n")
        self.thresholdFactorEntry.setValidator(qt.QDoubleValidator(self))
        layout.addWidget(self.thresholdFactorEntry, 2, 1)

        self.smoothStripGB = qt.QGroupBox("Apply smoothing prior to strip", self)
        self.smoothStripGB.setCheckable(True)
        self.smoothStripGB.setToolTip(
                "Apply a smoothing before subtracting strip background" +
                " in fit and estimate processes")
        smoothlayout = qt.QHBoxLayout(self.smoothStripGB)
        label = qt.QLabel("Smoothing width (Savitsky-Golay)")
        smoothlayout.addWidget(label)
        self.smoothingWidthSpin = qt.QSpinBox(self)
        self.smoothingWidthSpin.setToolTip(
            "Width parameter for Savitsky-Golay smoothing (number of samples, must be odd)")
        self.smoothingWidthSpin.setRange(3, 101)
        self.smoothingWidthSpin.setSingleStep(2)
        smoothlayout.addWidget(self.smoothingWidthSpin)

        layout.addWidget(self.smoothStripGB, 3, 0, 1, 2)

        layout.setRowStretch(4, 1)

        self.setDefault()

    def setDefault(self, default_dict=None):
        """Set default values for all widgets.

        :param default_dict: If a default config dictionary is provided as
            a parameter, its values are used as default values."""
        if default_dict is None:
            default_dict = {}

        self.setChecked(
                default_dict.get('StripBackgroundFlag', True))

        self.stripWidthSpin.setValue(
                default_dict.get('StripWidth', 2))
        self.numIterationsSpin.setValue(
                default_dict.get('StripIterations', 5000))
        self.thresholdFactorEntry.setText(
                str(default_dict.get('StripThreshold', 1.0)))
        self.smoothStripGB.setChecked(
                default_dict.get('SmoothingFlag', False))
        self.smoothingWidthSpin.setValue(
                default_dict.get('SmoothingWidth', 3))

    def get(self):
        """Return a dictionary of background subtraction parameters, to be
        processed by the :meth:`configure` method of the selected fit theory.
        """
        ddict = {
            'StripBackgroundFlag': self.isChecked(),
            'StripWidth': self.stripWidthSpin.value(),
            'StripIterations': self.numIterationsSpin.value(),
            'StripThreshold': safe_float(self.thresholdFactorEntry.text()),
            'SmoothingFlag': self.smoothStripGB.isChecked(),
            'SmoothingWidth': self.smoothingWidthSpin.value()
        }
        return ddict


def safe_float(string_, default=1.0):
    """Convert a string into a float.
    If the conversion fails, return the default value.
    """
    try:
        ret = float(string_)
    except ValueError:
        return default
    else:
        return ret


def safe_int(string_, default=1):
    """Convert a string into a integer.
    If the conversion fails, return the default value.
    """
    try:
        ret = int(float(string_))
    except ValueError:
        return default
    else:
        return ret


def getFitConfigDialog(parent=None, default=None, modal=True):
    """Instantiate and return a fit configuration dialog, adapted
    for configuring standard fit theories from
    :mod:`silx.math.fit.fittheories`.

    :return: Instance of :class:`TabsDialogData` with 3 tabs:
        :class:`ConstraintsPage`, :class:`SearchPage` and
        :class:`BackgroundPage`
    """
    tdd = TabsDialogData(parent=parent, default=default)
    tdd.addTab(ConstraintsPage(), label="Constraints")
    tdd.addTab(SearchPage(), label="Peak search")
    tdd.addTab(BackgroundPage(), label="Background")
    # apply default to newly added pages
    tdd.setDefault()

    return tdd


def main():
    a = qt.QApplication([])

    mw = qt.QMainWindow()
    mw.show()

    tdd = getFitConfigDialog(mw, default={"a": 1})
    tdd.show()
    tdd.exec()
    print("TabsDialogData result: ", tdd.result())
    print("TabsDialogData output: ", tdd.output)

    a.exec()

if __name__ == "__main__":
    main()



# /*##########################################################################
#
# Copyright (c) 2017-2021 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""This module provides 2D data and RGB(A) image item class.
"""

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "15/11/2017"

import numpy

from ..scene import primitives, utils
from .core import DataItem3D, ItemChangedType
from .mixins import ColormapMixIn, InterpolationMixIn
from ._pick import PickingResult


class _Image(DataItem3D, InterpolationMixIn):
    """Base class for images

    :param parent: The View widget this item belongs to.
    """

    def __init__(self, parent=None):
        DataItem3D.__init__(self, parent=parent)
        InterpolationMixIn.__init__(self)

    def _setPrimitive(self, primitive):
        InterpolationMixIn._setPrimitive(self, primitive)

    def getData(self, copy=True):
        raise NotImplementedError()

    def _pickFull(self, context):
        """Perform picking in this item at given widget position.

        :param PickContext context: Current picking context
        :return: Object holding the results or None
        :rtype: Union[None,PickingResult]
        """
        rayObject = context.getPickingSegment(frame=self._getScenePrimitive())
        if rayObject is None:
            return None

        points = utils.segmentPlaneIntersect(
            rayObject[0, :3],
            rayObject[1, :3],
            planeNorm=numpy.array((0., 0., 1.), dtype=numpy.float64),
            planePt=numpy.array((0., 0., 0.), dtype=numpy.float64))

        if len(points) == 1:  # Single intersection
            if points[0][0] < 0. or points[0][1] < 0.:
                return None  # Outside image
            row, column = int(points[0][1]), int(points[0][0])
            data = self.getData(copy=False)
            height, width = data.shape[:2]
            if row < height and column < width:
                return PickingResult(
                    self,
                    positions=[(points[0][0], points[0][1], 0.)],
                    indices=([row], [column]))
            else:
                return None  # Outside image
        else:  # Either no intersection or segment and image are coplanar
            return None


class ImageData(_Image, ColormapMixIn):
    """Description of a 2D image data.

    :param parent: The View widget this item belongs to.
    """

    def __init__(self, parent=None):
        _Image.__init__(self, parent=parent)
        ColormapMixIn.__init__(self)

        self._data = numpy.zeros((0, 0), dtype=numpy.float32)

        self._image = primitives.ImageData(self._data)
        self._getScenePrimitive().children.append(self._image)

        # Connect scene primitive to mix-in class
        ColormapMixIn._setSceneColormap(self, self._image.colormap)
        _Image._setPrimitive(self, self._image)

    def setData(self, data, copy=True):
        """Set the image data to display.

        The data will be casted to float32.

        :param numpy.ndarray data: The image data
        :param bool copy: True (default) to copy the data,
                          False to use as is (do not modify!).
        """
        self._image.setData(data, copy=copy)
        self._setColormappedData(self.getData(copy=False), copy=False)
        self._updated(ItemChangedType.DATA)

    def getData(self, copy=True):
        """Get the image data.

        :param bool copy:
            True (default) to get a copy,
            False to get internal representation (do not modify!).
        :rtype: numpy.ndarray
        :return: The image data
        """
        return self._image.getData(copy=copy)


class ImageRgba(_Image, InterpolationMixIn):
    """Description of a 2D data RGB(A) image.

    :param parent: The View widget this item belongs to.
    """

    def __init__(self, parent=None):
        _Image.__init__(self, parent=parent)
        InterpolationMixIn.__init__(self)

        self._data = numpy.zeros((0, 0, 3), dtype=numpy.float32)

        self._image = primitives.ImageRgba(self._data)
        self._getScenePrimitive().children.append(self._image)

        # Connect scene primitive to mix-in class
        _Image._setPrimitive(self, self._image)

    def setData(self, data, copy=True):
        """Set the RGB(A) image data to display.

        Supported array format: float32 in [0, 1], uint8.

        :param numpy.ndarray data:
            The RGBA image data as an array of shape (H, W, Channels)
        :param bool copy: True (default) to copy the data,
                          False to use as is (do not modify!).
        """
        self._image.setData(data, copy=copy)
        self._updated(ItemChangedType.DATA)

    def getData(self, copy=True):
        """Get the image data.

        :param bool copy:
            True (default) to get a copy,
            False to get internal representation (do not modify!).
        :rtype: numpy.ndarray
        :return: The image data
        """
        return self._image.getData(copy=copy)


class _HeightMap(DataItem3D):
    """Base class for 2D data array displayed as a height field.

    :param parent: The View widget this item belongs to.
    """

    def __init__(self, parent=None):
        DataItem3D.__init__(self, parent=parent)
        self.__data = numpy.zeros((0, 0), dtype=numpy.float32)

    def _pickFull(self, context, threshold=0., sort='depth'):
        """Perform picking in this item at given widget position.

        :param PickContext context: Current picking context
        :param float threshold: Picking threshold in pixel.
            Perform picking in a square of size threshold x threshold.
        :param str sort: How returned indices are sorted:

            - 'index' (default): sort by the value of the indices
            - 'depth':  Sort by the depth of the points from the current
              camera point of view.
        :return: Object holding the results or None
        :rtype: Union[None,PickingResult]
        """
        assert sort in ('index', 'depth')

        rayNdc = context.getPickingSegment(frame='ndc')
        if rayNdc is None:  # No picking outside viewport
            return None

        # TODO no colormapped or color data
        # Project data to NDC
        heightData = self.getData(copy=False)
        if heightData.size == 0:
            return  # Nothing displayed

        height, width = heightData.shape
        z = numpy.ravel(heightData)
        y, x = numpy.mgrid[0:height, 0:width]
        dataPoints = numpy.transpose((numpy.ravel(x),
                                      numpy.ravel(y),
                                      z,
                                      numpy.ones_like(z)))

        primitive = self._getScenePrimitive()

        pointsNdc = primitive.objectToNDCTransform.transformPoints(
            dataPoints, perspectiveDivide=True)

        # Perform picking
        distancesNdc = numpy.abs(pointsNdc[:, :2] - rayNdc[0, :2])
        # TODO issue with symbol size: using pixel instead of points
        threshold += 1.  # symbol size
        thresholdNdc = 2. * threshold / numpy.array(primitive.viewport.size)
        picked = numpy.where(numpy.logical_and(
                numpy.all(distancesNdc < thresholdNdc, axis=1),
                numpy.logical_and(rayNdc[0, 2] <= pointsNdc[:, 2],
                                  pointsNdc[:, 2] <= rayNdc[1, 2])))[0]

        if sort == 'depth':
            # Sort picked points from front to back
            picked = picked[numpy.argsort(pointsNdc[picked, 2])]

        if picked.size > 0:
            # Convert indices from 1D to 2D
            return PickingResult(self,
                                 positions=dataPoints[picked, :3],
                                 indices=(picked // width, picked % width),
                                 fetchdata=self.getData)
        else:
            return None

    def setData(self, data, copy: bool=True):
        """Set the height field data.

        :param data:
        :param copy: True (default) to copy the data,
            False to use as is (do not modify!).
        """
        data = numpy.array(data, copy=copy)
        assert data.ndim == 2

        self.__data = data
        self._updated(ItemChangedType.DATA)

    def getData(self, copy: bool=True) -> numpy.ndarray:
        """Get the height field 2D data.

        :param bool copy:
            True (default) to get a copy,
            False to get internal representation (do not modify!).
        """
        return numpy.array(self.__data, copy=copy)


class HeightMapData(_HeightMap, ColormapMixIn):
    """Description of a 2D height field associated to a colormapped dataset.

    :param parent: The View widget this item belongs to.
    """

    def __init__(self, parent=None):
        _HeightMap.__init__(self, parent=parent)
        ColormapMixIn.__init__(self)

        self.__data = numpy.zeros((0, 0), dtype=numpy.float32)

    def _updated(self, event=None):
        if event == ItemChangedType.DATA:
            self.__updateScene()
        super()._updated(event=event)

    def __updateScene(self):
        """Update display primitive to use"""
        self._getScenePrimitive().children = []  # Remove previous primitives
        ColormapMixIn._setSceneColormap(self, None)

        if not self.isVisible():
            return  # Update when visible

        data = self.getColormappedData(copy=False)
        heightData = self.getData(copy=False)

        if data.size == 0 or heightData.size == 0:
            return  # Nothing to display

        # Display as a set of points
        height, width = heightData.shape
        # Generates coordinates
        y, x = numpy.mgrid[0:height, 0:width]

        if data.shape != heightData.shape:  # data and height size miss-match
            # Colormapped data is interpolated (nearest-neighbour) to match the height field
            data = data[numpy.floor(y * data.shape[0] / height).astype(numpy.int32),
                        numpy.floor(x * data.shape[1] / height).astype(numpy.int32)]

        x = numpy.ravel(x)
        y = numpy.ravel(y)

        primitive = primitives.Points(
            x=x,
            y=y,
            z=numpy.ravel(heightData),
            value=numpy.ravel(data),
            size=1)
        primitive.marker = 's'
        ColormapMixIn._setSceneColormap(self, primitive.colormap)
        self._getScenePrimitive().children = [primitive]

    def setColormappedData(self, data, copy: bool=True):
        """Set the 2D data used to compute colors.

        :param data: 2D array of data
        :param copy: True (default) to copy the data,
            False to use as is (do not modify!).
        """
        data = numpy.array(data, copy=copy)
        assert data.ndim == 2

        self.__data = data
        self._updated(ItemChangedType.DATA)

    def getColormappedData(self, copy: bool=True) -> numpy.ndarray:
        """Returns the 2D data used to compute colors.

        :param copy:
            True (default) to get a copy,
            False to get internal representation (do not modify!).
        """
        return numpy.array(self.__data, copy=copy)


class HeightMapRGBA(_HeightMap):
    """Description of a 2D height field associated to a RGB(A) image.

    :param parent: The View widget this item belongs to.
    """

    def __init__(self, parent=None):
        _HeightMap.__init__(self, parent=parent)

        self.__rgba = numpy.zeros((0, 0, 3), dtype=numpy.float32)

    def _updated(self, event=None):
        if event == ItemChangedType.DATA:
            self.__updateScene()
        super()._updated(event=event)

    def __updateScene(self):
        """Update display primitive to use"""
        self._getScenePrimitive().children = []  # Remove previous primitives

        if not self.isVisible():
            return  # Update when visible

        rgba = self.getColorData(copy=False)
        heightData = self.getData(copy=False)
        if rgba.size == 0 or heightData.size == 0:
            return  # Nothing to display

        # Display as a set of points
        height, width = heightData.shape
        # Generates coordinates
        y, x = numpy.mgrid[0:height, 0:width]

        if rgba.shape[:2] != heightData.shape:  # image and height size miss-match
            # RGBA data is interpolated (nearest-neighbour) to match the height field
            rgba = rgba[numpy.floor(y * rgba.shape[0] / height).astype(numpy.int32),
                        numpy.floor(x * rgba.shape[1] / height).astype(numpy.int32)]

        x = numpy.ravel(x)
        y = numpy.ravel(y)

        primitive = primitives.ColorPoints(
            x=x,
            y=y,
            z=numpy.ravel(heightData),
            color=rgba.reshape(-1, rgba.shape[-1]),
            size=1)
        primitive.marker = 's'
        self._getScenePrimitive().children = [primitive]

    def setColorData(self, data, copy: bool=True):
        """Set the RGB(A) image to use.

        Supported array format: float32 in [0, 1], uint8.

        :param data:
            The RGBA image data as an array of shape (H, W, Channels)
        :param copy: True (default) to copy the data,
            False to use as is (do not modify!).
        """
        data = numpy.array(data, copy=copy)
        assert data.ndim == 3
        assert data.shape[-1] in (3, 4)
        # TODO check type

        self.__rgba = data
        self._updated(ItemChangedType.DATA)

    def getColorData(self, copy: bool=True) -> numpy.ndarray:
        """Get the RGB(A) image data.

        :param copy: True (default) to get a copy,
            False to get internal representation (do not modify!).
        """
        return numpy.array(self.__rgba, copy=copy)



# /*##########################################################################
#
# Copyright (c) 2020 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""This module provides the :func:`getQEventName` utility function."""

from silx.gui import qt


QT_EVENT_NAMES = {
    0: "None",
    114: "ActionAdded",
    113: "ActionChanged",
    115: "ActionRemoved",
    99: "ActivationChange",
    121: "ApplicationActivate",
    # ApplicationActivate: "ApplicationActivated",
    122: "ApplicationDeactivate",
    36: "ApplicationFontChange",
    37: "ApplicationLayoutDirectionChange",
    38: "ApplicationPaletteChange",
    214: "ApplicationStateChange",
    35: "ApplicationWindowIconChange",
    68: "ChildAdded",
    69: "ChildPolished",
    71: "ChildRemoved",
    40: "Clipboard",
    19: "Close",
    200: "CloseSoftwareInputPanel",
    178: "ContentsRectChange",
    82: "ContextMenu",
    183: "CursorChange",
    52: "DeferredDelete",
    60: "DragEnter",
    62: "DragLeave",
    61: "DragMove",
    63: "Drop",
    170: "DynamicPropertyChange",
    98: "EnabledChange",
    10: "Enter",
    150: "EnterEditFocus",
    124: "EnterWhatsThisMode",
    206: "Expose",
    116: "FileOpen",
    8: "FocusIn",
    9: "FocusOut",
    23: "FocusAboutToChange",
    97: "FontChange",
    198: "Gesture",
    202: "GestureOverride",
    188: "GrabKeyboard",
    186: "GrabMouse",
    159: "GraphicsSceneContextMenu",
    164: "GraphicsSceneDragEnter",
    166: "GraphicsSceneDragLeave",
    165: "GraphicsSceneDragMove",
    167: "GraphicsSceneDrop",
    163: "GraphicsSceneHelp",
    160: "GraphicsSceneHoverEnter",
    162: "GraphicsSceneHoverLeave",
    161: "GraphicsSceneHoverMove",
    158: "GraphicsSceneMouseDoubleClick",
    155: "GraphicsSceneMouseMove",
    156: "GraphicsSceneMousePress",
    157: "GraphicsSceneMouseRelease",
    182: "GraphicsSceneMove",
    181: "GraphicsSceneResize",
    168: "GraphicsSceneWheel",
    18: "Hide",
    27: "HideToParent",
    127: "HoverEnter",
    128: "HoverLeave",
    129: "HoverMove",
    96: "IconDrag",
    101: "IconTextChange",
    83: "InputMethod",
    207: "InputMethodQuery",
    169: "KeyboardLayoutChange",
    6: "KeyPress",
    7: "KeyRelease",
    89: "LanguageChange",
    90: "LayoutDirectionChange",
    76: "LayoutRequest",
    11: "Leave",
    151: "LeaveEditFocus",
    125: "LeaveWhatsThisMode",
    88: "LocaleChange",
    176: "NonClientAreaMouseButtonDblClick",
    174: "NonClientAreaMouseButtonPress",
    175: "NonClientAreaMouseButtonRelease",
    173: "NonClientAreaMouseMove",
    177: "MacSizeChange",
    43: "MetaCall",
    102: "ModifiedChange",
    4: "MouseButtonDblClick",
    2: "MouseButtonPress",
    3: "MouseButtonRelease",
    5: "MouseMove",
    109: "MouseTrackingChange",
    13: "Move",
    197: "NativeGesture",
    208: "OrientationChange",
    12: "Paint",
    39: "PaletteChange",
    131: "ParentAboutToChange",
    21: "ParentChange",
    212: "PlatformPanel",
    217: "PlatformSurface",
    75: "Polish",
    74: "PolishRequest",
    123: "QueryWhatsThis",
    106: "ReadOnlyChange",
    199: "RequestSoftwareInputPanel",
    14: "Resize",
    204: "ScrollPrepare",
    205: "Scroll",
    117: "Shortcut",
    51: "ShortcutOverride",
    17: "Show",
    26: "ShowToParent",
    50: "SockAct",
    192: "StateMachineSignal",
    193: "StateMachineWrapped",
    112: "StatusTip",
    100: "StyleChange",
    87: "TabletMove",
    92: "TabletPress",
    93: "TabletRelease",
    171: "TabletEnterProximity",
    172: "TabletLeaveProximity",
    219: "TabletTrackingChange",
    22: "ThreadChange",
    1: "Timer",
    120: "ToolBarChange",
    110: "ToolTip",
    184: "ToolTipChange",
    194: "TouchBegin",
    209: "TouchCancel",
    196: "TouchEnd",
    195: "TouchUpdate",
    189: "UngrabKeyboard",
    187: "UngrabMouse",
    78: "UpdateLater",
    77: "UpdateRequest",
    111: "WhatsThis",
    118: "WhatsThisClicked",
    31: "Wheel",
    132: "WinEventAct",
    24: "WindowActivate",
    103: "WindowBlocked",
    25: "WindowDeactivate",
    34: "WindowIconChange",
    105: "WindowStateChange",
    33: "WindowTitleChange",
    104: "WindowUnblocked",
    203: "WinIdChange",
    126: "ZOrderChange",
    65535: "MaxUser",
}


def getQEventName(eventType):
    """
    Returns the name of a QEvent.

    :param Union[int,qt.QEvent] eventType: A QEvent or a QEvent type.
    :returns: str
    """
    if isinstance(eventType, qt.QEvent):
        eventType = eventType.type()
    if 1000 <= eventType <= 65535:
        return "User_%d" % eventType
    name = QT_EVENT_NAMES.get(eventType, None)
    if name is not None:
        return name
    return "Unknown_%d" % eventType



#!/usr/bin/env python
# /*##########################################################################
#
# Copyright (c) 2018-2021 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Test of the FFT module"""

from os import path
import logging
import numpy as np
import unittest
from pkg_resources import parse_version
import pytest
from tempfile import TemporaryDirectory
try:
    from scipy.misc import ascent
    __have_scipy = True
except ImportError:
    __have_scipy = False
from silx.utils.testutils import ParametricTestCase
from silx.math.fft.fft import FFT
from silx.math.fft.clfft import __have_clfft__
from silx.math.fft.cufft import __have_cufft__
from silx.math.fft.fftw import __have_fftw__, import_wisdom, export_wisdom, get_wisdom_file

if __have_cufft__:
    import atexit
    import pycuda.driver as cuda
    from pycuda.tools import clear_context_caches

def get_cuda_context(device_id=None, cleanup_at_exit=True):
    """
    Create or get a CUDA context.
    """
    current_ctx = cuda.Context.get_current()
    # If a context already exists, use this one
    # TODO what if the device used is different from device_id ?
    if current_ctx is not None:
        return current_ctx
    # Otherwise create a new context
    cuda.init()

    if device_id is None:
        device_id = 0
    # Use the Context obtained by retaining the device's primary context,
    # which is the one used by the CUDA runtime API (ex. scikit-cuda).
    # Unlike Context.make_context(), the newly-created context is not made current.
    context = cuda.Device(device_id).retain_primary_context()
    context.push()
    # Register a clean-up function at exit
    def _finish_up(context):
        if context is not None:
            context.pop()
            context = None
        clear_context_caches()
    if cleanup_at_exit:
        atexit.register(_finish_up, context)
    return context

logger = logging.getLogger(__name__)

class TransformInfos(object):
    def __init__(self):
        self.dimensions = [
            "1D",
            "batched_1D",
            "2D",
            "batched_2D",
            "3D",
        ]
        self.modes = {
            "R2C": np.float32,
            "R2C_double": np.float64,
            "C2C": np.complex64,
            "C2C_double": np.complex128,
        }
        self.sizes = {
            "1D": [(128,), (127,)],
            "2D": [(128, 128), (128, 127), (127, 128), (127, 127)],
            "3D": [(64, 64, 64), (64, 64, 63), (64, 63, 64), (63, 64, 64),
                 (64, 63, 63), (63, 64, 63), (63, 63, 64), (63, 63, 63)]
        }
        self.axes = {
            "1D": None,
            "batched_1D": (-1,),
            "2D": None,
            "batched_2D": (-2, -1),
            "3D": None,
        }
        self.sizes["batched_1D"] = self.sizes["2D"]
        self.sizes["batched_2D"] = self.sizes["3D"]


class Data(object):
    def __init__(self):
        self.data = ascent().astype("float32")
        self.data1d = self.data[:, 0]  # non-contiguous data
        self.data3d = np.tile(self.data[:64, :64], (64, 1, 1))
        self.data_refs = {
            1: self.data1d,
            2: self.data,
            3: self.data3d,
        }


@unittest.skipUnless(__have_scipy, "scipy is missing")
@pytest.mark.usefixtures("test_options_class_attr")
class TestFFT(ParametricTestCase):
    """Test cuda/opencl/fftw backends of FFT"""

    def setUp(self):
        self.tol = {
            np.dtype("float32"): 1e-3,
            np.dtype("float64"): 1e-9,
            np.dtype("complex64"): 1e-3,
            np.dtype("complex128"): 1e-9,
        }
        self.transform_infos = TransformInfos()
        self.test_data = Data()

    @staticmethod
    def calc_mae(arr1, arr2):
        """
        Compute the Max Absolute Error between two arrays
        """
        return np.max(np.abs(arr1 - arr2))

    @unittest.skipIf(not __have_cufft__,
                     "cuda back-end requires pycuda and scikit-cuda")
    def test_cuda(self):
        get_cuda_context()

        # Error is higher when using cuda. fast_math mode ?
        self.tol[np.dtype("float32")] *= 2

        self.__run_tests(backend="cuda")

    @unittest.skipIf(not __have_clfft__,
                     "opencl back-end requires pyopencl and gpyfft")
    def test_opencl(self):
        from silx.opencl.common import ocl
        if ocl is not None:
            self.__run_tests(backend="opencl", ctx=ocl.create_context())

    @unittest.skipIf(not __have_fftw__,
                     "fftw back-end requires pyfftw")
    def test_fftw(self):
        self.__run_tests(backend="fftw")

    def __run_tests(self, backend, **extra_args):
        """Run all tests with the given backend

        :param str backend:
        :param dict extra_args: Additional arguments to provide to FFT
        """
        for trdim in self.transform_infos.dimensions:
            for mode in self.transform_infos.modes:
                for size in self.transform_infos.sizes[trdim]:
                    with self.subTest(trdim=trdim, mode=mode, size=size):
                        self.__test(backend, trdim, mode, size, **extra_args)

    def __test(self, backend, trdim, mode, size, **extra_args):
        """Compare given backend with numpy for given conditions"""
        logger.debug("backend: %s, trdim: %s, mode: %s, size: %s",
                     backend, trdim, mode, str(size))
        if size == "3D" and self.test_options.TEST_LOW_MEM:
            self.skipTest("low mem")

        ndim = len(size)
        input_data = self.test_data.data_refs[ndim].astype(
            self.transform_infos.modes[mode])
        tol = self.tol[np.dtype(input_data.dtype)]
        if trdim == "3D":
            tol *= 10  # Error is relatively high in high dimensions
        # It seems that cuda has problems with C2D batched 1D
        if trdim == "batched_1D" and backend == "cuda" and mode == "C2C":
            tol *= 10

        # Python < 3.5 does not want to mix **extra_args with existing kwargs
        fft_args = {
            "template": input_data,
            "axes": self.transform_infos.axes[trdim],
            "backend": backend,
        }
        fft_args.update(extra_args)
        F = FFT(
            **fft_args
        )
        F_np = FFT(
            template=input_data,
            axes=self.transform_infos.axes[trdim],
            backend="numpy"
        )

        # Forward FFT
        res = F.fft(input_data)
        res_np = F_np.fft(input_data)
        mae = self.calc_mae(res, res_np)
        all_close = np.allclose(res, res_np, atol=tol, rtol=tol),
        self.assertTrue(
            all_close,
            "FFT %s:%s, MAE(%s, numpy) = %f (tol = %.2e)" % (mode, trdim, backend, mae, tol)
        )

        # Inverse FFT
        res2 = F.ifft(res)
        mae = self.calc_mae(res2, input_data)
        self.assertTrue(
            mae < tol,
            "IFFT %s:%s, MAE(%s, numpy) = %f" % (mode, trdim, backend, mae)
        )


    # Test normalizations. silx FFT has three normalization modes:
    #    - "rescale" (default). FFT is unscaled, IFFT is scaled by 1/N.
    #      This corresponds to numpy normalize=None i.e normalize="backward"
    #    - "ortho": FFT/IFFT are both scaled with 1/sqrt(N) so that FFT is unitary.
    #    - "none": Neither FFT nor IFFT are not scaled, so IFFT(FFT(array)) = N*array

    norms_backends_support = {
        "numpy": {
            "supported_normalizations": ["rescale", "ortho", "none"],
        },
        "fftw": {
            "supported_normalizations": ["rescale", "ortho", "none"],
        },
        "opencl": {
            "supported_normalizations": ["rescale"],
        },
        "cuda": {
            "supported_normalizations": ["rescale", "none"],
        }
    }

    @staticmethod
    def _compute_numpy_normalized_fft(data, axes, silx_normalization_mode):
        if silx_normalization_mode in ["rescale", "none"]:
            return np.fft.rfftn(data, axes=axes, norm=None)
        elif silx_normalization_mode == "ortho":
            return np.fft.rfftn(data, axes=axes, norm="ortho")
        else:
            raise ValueError("Unknown normalization mode %s" % silx_normalization_mode)

    @staticmethod
    def _compute_numpy_normalized_ifft(data, axes, silx_normalization_mode):
        if silx_normalization_mode == "rescale":
            return np.fft.irfftn(data, axes=axes, norm=None)
        elif silx_normalization_mode == "ortho":
            return np.fft.irfftn(data, axes=axes, norm="ortho")
        elif silx_normalization_mode == "none":
            res =  np.fft.irfftn(data, axes=axes, norm=None)
            # This assumes a FFT on all the axes, won't work on batched FFT
            N = res.size
            return res * N
        else:
            raise ValueError("Unknown normalization mode %s" % silx_normalization_mode)

    @unittest.skipIf(not __have_fftw__, "fftw back-end requires pyfftw")
    def test_norms_fftw(self):
        return self._test_norms_with_backend("fftw")

    @unittest.skipIf(
        parse_version(np.version.version) <= parse_version("1.19.5"),
        "normalization does not work for numpy <= 1.19.5"
    )
    def test_norms_numpy(self):
        return self._test_norms_with_backend("numpy")

    @unittest.skipIf(not __have_clfft__, "opencl back-end requires pyopencl and gpyfft")
    def test_norms_opencl(self):
        from silx.opencl.common import ocl
        if ocl is not None:
            return self._test_norms_with_backend("opencl")

    @unittest.skipIf(not __have_cufft__, "cuda back-end requires pycuda and scikit-cuda")
    def test_norms_cuda(self):
        get_cuda_context()
        return self._test_norms_with_backend("cuda")

    def _test_norms_with_backend(self, backend_name):
        backend_params = self.norms_backends_support[backend_name]

        data = self.test_data.data
        tol = self.tol[np.dtype(data.dtype)]

        for norm in backend_params["supported_normalizations"]:
            fft = FFT(template=data, backend=backend_name, normalize=norm)
            res = fft.fft(data)
            ref = self._compute_numpy_normalized_fft(data, fft.axes, norm)
            assert np.allclose(res, ref, atol=tol, rtol=tol), "Something wrong with %s norm=%s" % (backend_name, norm)

            res2 = fft.ifft(res)
            ref2 = self._compute_numpy_normalized_ifft(ref, fft.axes, norm)
            # unscaled IFFT yields very large values. Use a relatively high "atol"
            assert np.allclose(res2, ref2, atol=res2.max()/1e6), "Something wrong with I%s norm=%s" % (backend_name, norm)


@unittest.skipUnless(__have_scipy, "scipy is missing")
class TestNumpyFFT(ParametricTestCase):
    """
    Test the Numpy backend individually.
    """

    def setUp(self):
        transforms = {
            "1D": {
                True: (np.fft.rfft, np.fft.irfft),
                False: (np.fft.fft, np.fft.ifft),
            },
            "2D": {
                True: (np.fft.rfft2, np.fft.irfft2),
                False: (np.fft.fft2, np.fft.ifft2),
            },
            "3D": {
                True: (np.fft.rfftn, np.fft.irfftn),
                False: (np.fft.fftn, np.fft.ifftn),
            },
        }
        transforms["batched_1D"] = transforms["1D"]
        transforms["batched_2D"] = transforms["2D"]
        self.transforms = transforms
        self.transform_infos = TransformInfos()
        self.test_data = Data()

    def test(self):
        """Test the numpy backend against native fft.

        Results should be exactly the same.
        """
        for trdim in self.transform_infos.dimensions:
            for mode in self.transform_infos.modes:
                for size in self.transform_infos.sizes[trdim]:
                    with self.subTest(trdim=trdim, mode=mode, size=size):
                        self.__test(trdim, mode, size)

    def __test(self, trdim, mode, size):
        logger.debug("trdim: %s, mode: %s, size: %s", trdim, mode, str(size))
        ndim = len(size)
        input_data = self.test_data.data_refs[ndim].astype(
            self.transform_infos.modes[mode])
        np_fft, np_ifft = self.transforms[trdim][np.isrealobj(input_data)]

        F = FFT(
            template=input_data,
            axes=self.transform_infos.axes[trdim],
            backend="numpy"
        )
        # Test FFT
        res = F.fft(input_data)
        ref = np_fft(input_data)
        self.assertTrue(np.allclose(res, ref))

        # Test IFFT
        res2 = F.ifft(res)
        ref2 = np_ifft(ref)
        self.assertTrue(np.allclose(res2, ref2))


@pytest.mark.skipif(not(__have_fftw__), reason="Need fftw/pyfftw for this test")
def test_fftw_wisdom():
    """
    Test FFTW wisdom import/export mechanism
    """

    assert path.isdir(path.dirname(get_wisdom_file())) # Default: tempdir.gettempdir()

    with TemporaryDirectory(prefix="test_fftw_wisdom") as dname:
        subdir = path.join(dname, "subdir")
        get_wisdom_file(directory=subdir, create_dirs=False)
        assert not(path.isdir(subdir))
        fname = get_wisdom_file(directory=subdir, create_dirs=True)
        assert path.isdir(subdir)
        export_wisdom(fname)
        assert path.isfile(fname)
        import_wisdom(fname)

